/*
 * attr.c		Item Management
 *
 * Copyright (c) 2001-2005 Thomas Graf <tgraf@suug.ch>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <bmon/bmon.h>
#include <bmon/attr.h>
#include <bmon/unit.h>
#include <bmon/utils.h>

#define ID_HASHSIZ 1024
#define NAME_HASHSIZ 32

struct attr_map attrs[] = {
	[BYTES] = { .am_name = "bytes" },
	[PACKETS] = { .am_name = "packets" },
	[ERRORS] = { .am_name "errors" },
	[DROP] = { .am_name = "drop" },
	[FIFO] = { .am_name = "fifo" },
	[FRAME] = { .am_name = "frame" },
	[COMPRESSED] = { .am_name = "compressed" },
	[MULTICAST] = { .am_name = "multicast" },
	[OVERLIMITS] = { .am_name = "overlimits" },
	[BPS] = { .am_name = "bps" },
	[PPS] = { .am_name = "pps" },
	[QLEN] = { .am_name = "qlen" },
	[BACKLOG] = { .am_name = "backlog" },
	[REQUEUES] = { .am_name = "requeues" },
	[COLLISIONS] = { .am_name = "collisions" },
	[LENGTH_ERRORS] = { .am_name = "len_err" },
	[OVER_ERRORS] = { .am_name = "over_err" },
	[CRC_ERRORS] = { .am_name = "crc_err" },
	[MISSED_ERRORS] = { .am_name = "missed_err" },
	[ABORTED_ERRORS] = { .am_name = "abort_err" },
	[HEARTBEAT_ERRORS] = { .am_name = "hbeat_err" },
	[WINDOW_ERRORS] = { .am_name = "window_err" },
	[CARRIER_ERRORS] = { .am_name = "carrier_err" }
};

static struct attr_type *attr_ht_id[ID_HASHSIZ];
static struct attr_type *attr_ht_name[NAME_HASHSIZ];
static int attr_gen = 1;

static inline int attr_hash_id(int attr_id)
{
	return attr_id & (ID_HASHSIZ-1);
}

static inline int attr_hash_name(const char *name)
{
	return (*name - 0x61) & (NAME_HASHSIZ-1);
}

struct attr_type *lookup_attrtype(const char *name)
{
	struct attr_type *t;
	int hash = attr_hash_name(name);

	for (t = attr_ht_name[hash]; t; t = t->at_next_name)
		if (!strcmp(name, t->at_name))
			return t;

	return NULL;
}

struct attr_type *lookup_attrtype_byid(int id)
{
	struct attr_type *t;
	int hash = attr_hash_id(id);

	for (t = attr_ht_id[hash]; t; t = t->at_next_id)
		if (id == t->at_id)
			return t;

	return NULL;
}

int attr_resolve_map(struct attr_map *attrs, int len)
{
	int i;

	for (i = 0; i < len; i++) {
		struct attr_map *m = &attrs[i];
		m->am_id = attr_name2type(m->am_name);
		if (m->am_id == INT_MAX) {
			fprintf(stderr, "Cannot resolve attribute %s\n",
				m->am_name);
			return -1;
		}
	}

	return 0;
}

int attr_resolve_global_map(void)
{
	if (attr_resolve_map(attrs, ARRAY_SIZE(attrs)) < 0)
		return -1;

	return 0;
}
	

int attr_name2type(const char *name)
{
	struct attr_type *t = lookup_attrtype(name);
	return t ? t->at_id : INT_MAX;
}

const char *attr_type2name(int id)
{
	struct attr_type *t = lookup_attrtype_byid(id);
	return t ? t->at_name : NULL;
}

const char *attr_type2desc(int id)
{
	static char str[256];
	struct attr_type *t = lookup_attrtype_byid(id);

	if (t)
		return t->at_desc;

	snprintf(str, sizeof(str), "unknown (%d)", id);
	return str;
}

static inline int __check_flag(int id, int flags)
{
	struct attr_type *t = lookup_attrtype_byid(id);

	if (t && t->at_flags & flags)
		return 1;

	return 0;
}

int attr_ignore_overflows(int id)
{
	return __check_flag(id, ATTR_TYPE_IGN_OFLOWS);
}

int attr_is_rate(int id)
{
	return __check_flag(id, ATTR_TYPE_IS_RATE);
}

int attr_is_usage(int id)
{
	return __check_flag(id, ATTR_TYPE_IS_USAGE);
}

struct unit *attr_unit(int id)
{
	struct attr_type *t = lookup_attrtype_byid(id);
	return t ? t->at_unit : NULL;
}

int foreach_attr_type(int (*cb)(struct attr_type *, void *), void *arg)
{
	int i, err = 0;

	for (i = 0; i < NAME_HASHSIZ; i++) {
		struct attr_type *t;
		for (t = attr_ht_name[i]; t; t = t->at_next_name) {
			err = cb(t, arg);
			if (err < 0)
				break;
		}
	}

	return err;
}

void add_attribute(const char *name, const char *desc, const char *unit,
		   const char *flags)
{
	int id_hash, name_hash, fl = 0;
	struct attr_type *t;
	struct unit *u = lookup_unit(unit);

	if (u == NULL)
		quit("Unknown unit \"%s\" for attribute \"%s\"\n", unit, name);

	if (flags) {
		char *p, *o, *f = strdup(flags);
		o = f;
		do {
			p = strchr(o, ',');
			if (p) {
				*p = '\0';
				p++;
			}

			if (!strcasecmp(o, "ignoflows"))
				fl |= ATTR_TYPE_IGN_OFLOWS;
			else if (!strcasecmp(o, "israte"))
				fl |= (ATTR_TYPE_IS_RATE|ATTR_TYPE_IGN_OFLOWS);
			else if (!strcasecmp(o, "isusage"))
				fl |= ATTR_TYPE_IS_USAGE;

			o = p;
		} while (p);

		free(f);
	}

	t = xcalloc(1, sizeof(*t));

	t->at_id = attr_gen++;
	t->at_name = strdup(name);
	t->at_desc = strdup(desc);
	t->at_unit = u;
	t->at_flags = fl;

	id_hash = attr_hash_id(t->at_id);
	t->at_next_id = attr_ht_id[id_hash];
	attr_ht_id[id_hash] = t;

	name_hash = attr_hash_name(t->at_name);
	t->at_next_name = attr_ht_name[name_hash];
	attr_ht_name[name_hash] = t;

	DBG(1, "[DBG] New attribute %s \"%s\" unit %s %d\n",
	    t->at_name, t->at_desc, t->at_unit->u_name, t->at_flags);
}
