/*
 * src/bmon.c		   Bandwidth Monitor
 *
 * Copyright (c) 2001-2005 Thomas Graf <tgraf@suug.ch>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <bmon/bmon.h>
#include <bmon/conf.h>
#include <bmon/item.h>
#include <bmon/attr.h>
#include <bmon/itemtab.h>
#include <bmon/node.h>
#include <bmon/pipe.h>
#include <bmon/input.h>
#include <bmon/output.h>
#include <bmon/signal.h>
#include <bmon/utils.h>

int start_time;
int do_quit = 0;
int is_daemon = 0;

struct reader_timing rtiming;
static char *usage_text =
"Usage: bmon [OPTION]...\n" \
"\n" \
"Options:\n" \
"Startup:\n" \
"   -i, --input=MODPARM             Primary input module\n" \
"   -o, --output=MODPARM            Primary ouptut module\n" \
"   -I, --secondary-input=MODPARM   Secondary input modules\n" \
"   -O, --secondary-output=MODPARM  Secondary output modules\n" \
"   -f, --configfile=PATH           Alternative path to configuration file\n" \
"   -F, --configbasedir=PATH        Base path to configuration directory\n" \
"   -b, --after-reboot              Inform that a reboot has happened\n" \
"   -B, --reboot-file=PATH          Path to reboot detection file\n" \
"   -w, --wait-for-signal           Signal driven output intervals\n" \
"   -S, --send-signal=PID           Send SIGUSR1 to a running bmon instance\n" \
"   -d, --daemon                    Run as a daemon\n" \
"   -P, --pidfile=PATH              Path to the pidfile\n" \
"   -u, --uid=UID                   Drop privileges and change UID\n" \
"   -g, --gid=GID                   Drop privileges and change GID\n" \
"   -h, --help                      show this help text\n" \
"   -V, --version                   show version\n" \
"\n" \
"Input:\n" \
"   -p, --policy=POLICY             Interface acceptance policy\n" \
"   -a, --show-all                  Accept interfaces even if they are down\n" \
"   -A, --attr-policy=ATTRS         Attributes history configuration\n" \
"   -r, --read-interval=FLOAT       Read interval in seconds\n" \
"   -H, --heartbeat=HBEAT           Heartbeat factor (0..1)\n" \
"   -L, --lifetime=LIFETIME         Lifetime of a item in seconds\n" \
"   -t, --itemtab=PATH              Alternative path to itemtab file\n" \
"   -s, --sleep-interval=FLOAT      Sleep time in seconds\n" \
"\n" \
"Output:\n" \
"   -c, --use-si                    Use SI units\n" \
"   -N, --num-graphs=NUM            Number of graphs to draw\n" \
"\n" \
"Rate Estimation:\n" \
"   -R, --rate-interval=FLOAT       Rate interval in seconds\n" \
"\n" \
"Module configuration:\n" \
"   modparm := MODULE:optlist,MODULE:optlist,...\n" \
"   optlist := option;option;...\n" \
"   option  := TYPE[=VALUE]\n" \
"\n" \
"   Examples:\n" \
"       -O html:path=/var/www/html,distribution:port=2444;debug\n" \
"       -O list          # Shows a list of available modules\n" \
"       -O html:help     # Shows a help text for html module\n" \
"\n" \
"Interface selection:\n" \
"   policy  := [!]simple_regexp,[!]simple_regexp,...\n" \
"\n" \
"   Example: -p 'eth*,lo*,!eth1'\n" \
"\n" \
"Attributes selection:\n" \
"   attrs   := [!]name,[!]name,...\n" \
"\n" \
"   Example: -A !hbeat_err\n" \
"\n" \
"Please see the bmon(1) man pages for full documentation.\n";

static char *uid, *gid;

static void do_shutdown(void)
{
	static int done;
	
	if (!done) {
		done = 1;
		input_shutdown();
		output_shutdown();
	}
}

RETSIGTYPE sig_int(int unused)
{
	if (do_quit)
		exit(-1);
	do_quit = 1;
}

void sig_exit(void)
{
	do_shutdown();
}

void quit(const char *fmt, ...)
{
	static int done;
	va_list args;

	if (!done) {
		done = 1;
		do_shutdown();
	}

	va_start(args, fmt);
	if (is_daemon)
		vsyslog(LOG_ERR, fmt, args);
	else
		vfprintf(stderr, fmt, args);
	va_end(args);

	exit(1);
}

inline void xwarn(const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);
}

static inline void print_version(void)
{
	printf("bmon %s\n", PACKAGE_VERSION);
	printf("Copyright (C) 2001-2005 by Thomas Graf <tgraf@suug.ch>\n");
	printf("bmon comes with ABSOLUTELY NO WARRANTY. This is free " \
	       "software, and you\nare welcome to redistribute it under " \
	       "certain conditions. See the source\ncode for details.\n");
}

static inline void check_for_reboot(void)
{
	char *path = get_reboot_file();

	if (access(path, R_OK) == 0) {
		set_reboot(1);

		if (unlink(path) < 0)
			xwarn("Could not unlink reboot file: %s\n",
			      strerror(errno));
	}
}
	

static void parse_args(int argc, char *argv[])
{
	for (;;)
	{
		char *gostr = "hVf:i:I:o:O:p:r:s:S:P:wadA:cN:" \
			      "u:g:H:R:L:t:bB:F:";

#ifdef HAVE_GETOPT_LONG
		struct option long_opts[] = {
			{"help", 0, 0, 'h'},
			{"version", 0, 0, 'v'},
			{"configfile", 1, 0, 'f'},
			{"configbasedir", 1, 0, 'F'},
			{"input", 1, 0, 'i'},
			{"secondary-input", 1, 0, 'I'},
			{"output", 1, 0, 'o'},
			{"secondary-output", 1, 0, 'O'},
			{"policy", 1, 0, 'p'},
			{"read-interval", 1, 0, 'r'},
			{"sleep-interval", 1, 0, 's'},
			{"rate-interval", 1, 0, 'R'},
			{"send-signal", 1, 0, 'S'},
			{"pidfile", 1, 0, 'P'},
			{"after-reboot", 0, 0, 'b'},
			{"reboot-file", 1, 0, 'B'},
			{"wait-for-signal", 0, 0, 'w'},
			{"show-all", 0, 0, 'a'},
			{"daemon", 0, 0, 'd'},
			{"attr-policy", 1, 0, 'A'},
			{"use-si", 0, 0, 'c'},
			{"num-graphs", 1, 0, 'N'},
			{"uid", 1, 0, 'u'},
			{"gid", 1, 0, 'g'},
			{"heartbeat", 1, 0, 'H'},
			{"lifetime", 1, 0, 'L'},
			{"itemtab", 1, 0, 't'},
			{0, 0, 0, 0},
		};
		int c = getopt_long(argc, argv, gostr, long_opts, NULL);
#else
		int c = getopt(argc, argv, gostr;
#endif

		if (c == -1)
			break;
		
		switch (c)
		{
			case 'i':
				set_input(optarg);
				break;

			case 'I':
				set_sec_input(optarg);
				break;

			case 'o':
				set_output(optarg);
				break;

			case 'O':
				set_sec_output(optarg);
				break;
				
			case 'f':
				set_configfile(optarg);
				break;

			case 'F':
				set_configbasedir(optarg);
				break;

			case 'p':
				item_parse_policy(optarg);
				break;

			case 'A':
				parse_attr_policy(optarg);
				break;

			case 'r':
				set_read_interval(optarg);
				break;

			case 'R':
				set_rate_interval(optarg);
				break;

			case 'L':
				set_lifetime(optarg);
				break;

			case 's':
				set_sleep_time(optarg);
				break;

			case 'S':
				send_signal(optarg);
				exit(0);

			case 'w':
				set_signal_output(1);
				break;

			case 'a':
				set_show_only_running(0);
				break;

			case 'h':
				print_version();
				printf("\n%s", usage_text);
				exit(1);

			case 'N':
				set_ngraphs(strtol(optarg, NULL, 0));
				break;

			case 'c':
				set_use_si();
				break;

			case 'd':
				set_run_as_daemon(1);
				break;

			case 'P':
				set_pidfile(optarg);
				break;

			case 'b':
				set_reboot(1);
				break;

			case 'B':
				set_reboot_file(optarg);
				break;
			
			case 'u':
				uid = strdup(optarg);
				break;

			case 'g':
				gid = strdup(optarg);
				break;

			case 'H':
				set_hb_factor(optarg);
				break;

			case 't':
				set_itemtab(optarg);
				break;

			case 'V':
			case 'v':
				print_version();
				exit(0);

			default:
				quit("Aborting...\n");
		}
	}
}

static void calc_variance(timestamp_t *c, timestamp_t *ri)
{
	float v = (ts_to_float(c) / ts_to_float(ri)) * 100.0f;

	rtiming.rt_variance.v_error = v;
	rtiming.rt_variance.v_total += v;

	if (v > rtiming.rt_variance.v_max)
		rtiming.rt_variance.v_max = v;

	if (v < rtiming.rt_variance.v_min)
		rtiming.rt_variance.v_min = v;
}

static void drop_privs(void)
{
	struct passwd *uentry = NULL;
	struct group *gentry = NULL;

	if (gid)
		gentry = getgrnam(gid);

	if (uid)
		uentry = getpwnam(uid);

	if (gentry)
		if (setgid(gentry->gr_gid) < 0)
			quit("Unable to set group id %d: %s\n",
			    gentry->gr_gid, strerror(errno));

	if (uentry)
		if (setuid(uentry->pw_uid) < 0)
			quit("Unable to set user id %d: %s\n",
			    uentry->pw_uid, strerror(errno));
}

static void daemonize(void)
{
#ifdef HAVE_DAEMON 
	if (!daemon(0, 0))
		quit("Can't fork: %s\n", strerror(errno));
#else
	int devnull;
	pid_t pid;

	devnull = open("/dev/null", O_RDWR);

	if (devnull < 0)
		quit("Unable to open /dev/null: %s\n", strerror(errno));
		
	/* reopen stdin/stdout/stderr */
	if (dup2(devnull, 0) < 0) 
		quit("Can't dup /dev/null to STDIN: %s\n", strerror(errno));

	if (dup2(devnull, 1) < 0)  
		quit("Can't dup /dev/null to STDOUT: %s\n", strerror(errno));

	if (dup2(devnull, 2) < 0) 
		quit("Can't dup /dev/null to STDERR: %s\n", strerror(errno));

	if (chdir("/") < 0) 
		quit("Can't change into the root directory: %s\n",
		     strerror(errno));

	pid = fork();
	if (pid < 0) 
		quit("Can't fork: %s\n", strerror(errno));

	if (pid)  /* parent */
		exit(0);
	setsid();
#endif
	is_daemon = 1;
}

static void write_pidfile(void)
{
	FILE *pidfile;
	const char * pidfilepath = get_pidfile();

	pidfile = fopen(pidfilepath, "w");

	if (pidfile == NULL)
		quit("Can't open pidfile %s: %s\n", pidfilepath,
		     strerror(errno));

	fprintf(pidfile, "%i\n", getpid());


	if (fclose(pidfile) < 0)
		quit("Can't close pidfile %s: %s\n", pidfilepath,
		     strerror(errno));
}

void init_syslog(void)
{
	openlog("bmon", LOG_CONS | LOG_PID, LOG_DAEMON);
}

int main(int argc, char *argv[])
{
	unsigned long c_sleep_time;
	float read_interval;

	start_time = time(0);

	parse_args(argc, argv);
	read_configfile();
	read_itemtab();

	if (attr_resolve_global_map() < 0)
		quit("Could not resolve all attributes, missing " \
		     "configuration file?\n");

	check_for_reboot();

	c_sleep_time = get_sleep_time();
	read_interval = get_read_interval();

	if (((double) c_sleep_time/1000000.0f) > read_interval)
		c_sleep_time = (unsigned long) (read_interval * 1000000.0f);

	pipe_start();

	input_init();
	output_init();

	if (get_run_as_daemon()) {
		init_syslog();
		daemonize();
		write_pidfile();
	}

	drop_privs();

	{
		/*
		 * E  := Elapsed time
		 * NR := Next Read
		 * LR := Last Read
		 * RI := Read Interval
		 * ST := Sleep Time
		 * C  := Correction
		 */
		timestamp_t e, ri, tmp;
		unsigned long st;

		get_read_interval_as_ts(&ri);

		/*
		 * E := NOW()
		 */
		update_ts(&e);
		
		/*
		 * NR := E
		 */
		COPY_TS(&rtiming.rt_next_read, &e);
		
		for (;;) {
			output_pre();

			/*
			 * read the chucka chucka pipe
			 */
			pipe_handle();

			/*
			 * E := NOW()
			 */
			update_ts(&e);

			/*
			 * IF NR <= E THEN
			 */
			if (ts_le(&rtiming.rt_next_read, &e)) {
				timestamp_t c;

				/*
				 * C :=  (NR - E)
				 */
				ts_sub(&c, &rtiming.rt_next_read, &e);

				calc_variance(&c, &ri);

				/*
				 * LR := E
				 */
				COPY_TS(&rtiming.rt_last_read, &e);

				/*
				 * NR := E + RI + C
				 */
				ts_add(&rtiming.rt_next_read, &e, &ri);
				ts_add(&rtiming.rt_next_read,
				       &rtiming.rt_next_read, &c);
				
				reset_nodes();
				pipe_read();
				input_read();
				remove_unused_node_items();

				output_draw();

				output_post();

				/* Disable special paths for after reboot after
				 * the first run through all modules. */
				set_reboot(0);
			}

			if (do_quit)
				exit(0);

			if (got_resized())
				output_resize();

			/*
			 * ST := Configured ST
			 */
			st = c_sleep_time;

			/*
			 * IF (NR - E) < ST THEN
			 */
			ts_sub(&tmp, &rtiming.rt_next_read, &e);

			if (tmp.tv_sec < 0)
				continue;

			if (tmp.tv_sec == 0 && tmp.tv_usec < st) {
				if (tmp.tv_usec < 0)
					continue;
				/*
				 * ST := (NR - E)
				 */
				st = tmp.tv_usec;
			}
			
			/*
			 * SLEEP(ST)
			 */
			usleep(st);
		}
	}
	
	return 0; /* buddha says i'll never be reached */
}

static void __init bmon_init(void)
{
	memset(&rtiming, 0, sizeof(rtiming));
	rtiming.rt_variance.v_min = 10000000.0f;
	atexit(&sig_exit);
	signal(SIGINT, &sig_int);
}
