/*
 * bmondb.c		Database Maintainance Tool
 *
 * Copyright (c) 2001-2005 Thomas Graf <tgraf@suug.ch>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <bmon/bmon.h>
#include <bmon/node.h>
#include <bmon/output.h>
#include <bmon/graph.h>
#include <bmon/input.h>
#include <bmon/utils.h>
#include <bmon/db.h>

struct reader_timing rtiming;
static PGconn *db_conn;

static char *usage_text =
"Usage: bmondb [OPTION] ACTION...\n" \
"\n" \
"Options:\n" \
"   -H, --host=TEXT       Database host\n" \
"   -u, --user=TEXT       Database user\n" \
"   -n, --dbname=TEXT     Database name\n" \
"   -p, --password=TEXT   Database password\n" \
"   -h, --help            Show this help text\n" \
"   -v, --version         Show bmondb version\n" \
"\n" \
"Actions:\n" \
"\n" \
"  create\n" \
"    Create tables and install attribute descriptions\n" \
"\n" \
"  destroy\n" \
"    Destroy database and drop all tables.\n"
"\n" \
"  reinstall_rules\n" \
"    Reinstall database rules\n" \
"\n" \
"  analyze\n" \
"    Analyze database data to improve optimizations of queries\n" \
"\n" \
"  dump [--id]\n" \
"    Dump content of database for manual sanity check.\n" \
"\n" \
"  get [--from TS] [--to TS] [FILTER]\n" \
"    Lookup database content\n" \
"\n" \
"  FILTER := node[.interface[.attribute][@unit]]\n" \
"  ... where '*' represents a wildcard\n" \
"            {r | s | m | h | d} are valid units\n" \
"\n" \
"Please see the bmondb(1) man pages for full documentation.\n";

static void print_version(void)
{
	printf("bmondb %s\n", PACKAGE_VERSION);
	printf("Copyright (C) 2001-2005 by Thomas Graf <tgraf@suug.ch>\n");
	printf("bmon comes with ABSOLUTELY NO WARRANTY. This is free " \
	       "software, and you\nare welcome to redistribute it under " \
	       "certain conditions. See the source\ncode for details.\n");
}

void quit(const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);

	exit(1);
}

void xwarn(const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);
}


void diff_dates(xdate_t *a, xdate_t *b, timestamp_t *dst)
{
	timestamp_t at = {
		.tv_sec = mktime(&a->d_tm),
		.tv_usec = a->d_usec
	};
	timestamp_t bt = {
		.tv_sec = mktime(&b->d_tm),
		.tv_usec = b->d_usec
	};

	ts_sub(dst, &at, &bt);
}

static int do_create_tables(void)
{
	if (bdb_create_tables(db_conn) < 0)
		return -1;

	if (bdb_insert_attr_descs(db_conn) < 0)
		return -1;

	return 0;
}

static int do_destroy(void)
{
	if (bdb_drop_rules(db_conn) < 0)
		return -1;

	if (bdb_drop_tables(db_conn) < 0)
		return -1;

	return 0;
}

static int do_analyze(void)
{
	bdb_analyze(db_conn);

	return 0;
}

static inline int in_history_mode(struct db_filter *filter,
				  struct db_hist_params *params)
{
	return params || filter->f_field;
}

static int do_dump(int argc, char *argv[])
{
	struct db_node *node_cache;
	struct db_item *item_cache;
	struct db_attr *attr_cache;
	int i, n, m, print_id = 0;
	int node_cache_size, item_cache_size, attr_cache_size;

	if (bdb_update_caches(db_conn) < 0)
		return -1;

	for (;;) {
		const char *gostr = "i";
#ifdef HAVE_GETOPT_LONG
		struct option long_opts[] = {
			{"id", 0, 0, 'i'},
			{0, 0, 0, 0},
		};
		int c = getopt_long(argc, argv, gostr, long_opts, NULL);
#else
		int c = getopt(argc, argv, gostr);
#endif
		if (c == -1)
			break;
		else if (c == 'i')
			print_id = 1;
		else
			return -1;
	}

	node_cache = bdb_get_node_cache();
	node_cache_size = bdb_get_node_cache_rows();
	item_cache = bdb_get_item_cache();
	item_cache_size = bdb_get_item_cache_rows();
	attr_cache = bdb_get_attr_cache();
	attr_cache_size = bdb_get_attr_cache_rows();

	for (i = 0; i < node_cache_size; i++) {
		struct db_node *node = &node_cache[i];

		if (print_id)
			printf("Node %s: %d\n", node->name, node->db_id);
		else
			printf("%s: (idx: %04d dbid: %04d)\n",
			       node->name, node->index, node->db_id);

		for (n = 0; n < item_cache_size; n++) {
			struct db_item *item = &item_cache[n];

			if (node->db_id != item->node)
				continue;

			if (print_id)
				printf("  Item %s: %d.%d\n",
				       item->name, node->db_id, item->db_id);
			else
				printf("  %-14s idx: %04d dbid: %04d " \
				       "handle: 0x%04x parent: %04d " \
				       "node: %04d\n",
				       item->name, item->index, item->db_id,
				       item->handle, item->parent, item->node);

			for (m = 0; m < attr_cache_size; m++) {
				struct db_attr *attr = &attr_cache[m];

				if (item->db_id != attr->item)
					continue;

				if (print_id)
					printf("    Attribute %s: %d.%d.%d\n",
					       attr->name, node->db_id,
					       item->db_id, attr->db_id);
				else
					printf("    %-14s idx: %04d dbid: " \
					       "%04d item: %04d\n",
					       attr->name, attr->index,
					       attr->db_id, attr->item);
			}
		}
	}
	
	return 0;
}

static inline char *prepare_nodename(struct db_node_full *node, char *buf,
				      size_t len)
{
	memset(buf, 0, len);

	snprintf(buf, len, "%s", node->name);

	if (node->source[0]) {
		strncat(buf, " (", len - strlen(buf) - 1);
		strncat(buf, node->source, len - strlen(buf) - 1);
		strncat(buf, ")", len - strlen(buf) - 1);
	}

	return buf;
}

static inline char * prepare_itemname(struct db_item_full *item, char *buf,
				      size_t len)
{
	memset(buf, 0, len);

	if (item->indent) {
		int indent = item->indent > 10 ? 10 : item->indent;
		memset(buf, ' ', indent*2);
	}

	strncat(buf, item->name, len - strlen(buf) - 1);
	if (item->desc[0]) {
		strncat(buf, " (", len - strlen(buf) - 1);
		strncat(buf, item->desc, len - strlen(buf) - 1);
		strncat(buf, ")", len - strlen(buf) - 1);
	}

	return buf;
}

static void dump_item(struct db_item_full *item, struct db_collection *items)
{
	char name[64];
	struct db_item_full *p;

	printf("  %-30s",
	       prepare_itemname(item, name, sizeof(name)));

	if (item->rx_usage == -1)
		printf("  N/A");
	else
		printf(" %3d%%", item->rx_usage);

	if (item->tx_usage == -1)
		printf("  N/A");
	else
		printf(" %3d%%", item->tx_usage);

	printf("  %s\n", item->last_update);

	for_each_entry(p, items, struct db_item_full)
		if (p->parent == item->db_id)
			dump_item(p, items);
}


static int dump_node(struct db_node_full *node, struct db_filter *filter,
		     struct db_hist_params *params)
{
	char name[64];

	struct db_collection items;
	struct db_item_full *item;

	printf("%s\n", prepare_nodename(node, name, sizeof(name)));

	if (bdb_fetch_items(db_conn, node->db_id, NULL, &items) < 0)
		return -1;

	for_each_entry(item, &items, struct db_item_full) {
		/* skip over childs, we dump them recursively */
		if (item->parent >= 0)
			continue;

		dump_item(item, &items);
	}

	bdb_destroy_collection(&items);

	return 0;
}

static int do_get_node(struct db_filter *filter, struct db_hist_params *params)
{
	struct db_collection nodes;
	struct db_node_full *node;

	if (bdb_fetch_nodes(db_conn, filter->f_node, &nodes) < 0)
		return -1;

	printf("Name                               RX   TX  Last Update\n");
	printf("--------------------------------------------------------" \
	       "--------------\n");

	for_each_entry(node, &nodes, struct db_node_full)
		dump_node(node, filter, params);

	bdb_destroy_collection(&nodes);
	
	return 0;
}

static int print_attr_hist(struct db_item_full *item, struct db_filter *filter,
			   struct db_hist_params *params,
			   struct db_token *attr_filter)
{
	struct db_collection attrs;
	struct db_attr_full *attr;
	int no_deltas;
	int rxp, txp, drxp, dtxp, rxrp, txrp;
	char *rxu, *txu, *drxu, *dtxu, *rxru, *txru, *last_attr = NULL;
	double rx, tx, drx = 0.0f, dtx = 0.0f, rxr = 0.0f, txr = 0.0f;
	xdate_t cur_date, last_date;
	timestamp_t delta_time;
	b_cnt_t last_rx = 0, last_tx = 0, delta_rx, delta_tx;
	b_cnt_t rx_rate, tx_rate;

	memset(&last_date, 0, sizeof(last_date));

	if (bdb_fetch_attr_hist(db_conn, item->db_id, attr_filter, &attrs,
				params, filter->f_field) < 0)
		return -1;

	for_each_entry(attr, &attrs, struct db_attr_full) {
		if (!last_attr || strcmp(last_attr, attr->name)) {
			last_rx = last_tx = 0;
			no_deltas = 1;
		} else
			no_deltas = 0;

		if (bdb_parse_date(attr->hts, &cur_date) < 0)
			printf("      Error: Invalid date \"%s\"\n", attr->hts);

		if (attr->rx_cnt < last_rx) {
			printf("      Warning: RX counter rebound\n");
			last_rx = 0;
		}

		if (attr->tx_cnt < last_tx) {
			printf("      Warning: TX counter rebound\n");
			last_tx = 0;
		}

		if (!no_deltas) {
			diff_dates(&cur_date, &last_date, &delta_time);

			delta_rx = attr->rx_cnt - last_rx;
			drx = cancel_down(delta_rx, attr->unit, &drxu, &drxp);

			delta_tx = attr->tx_cnt - last_tx;
			dtx = cancel_down(delta_tx, attr->unit, &dtxu, &dtxp);

			rx_rate = delta_rx / ts_to_float(&delta_time);
			rxr = cancel_down(rx_rate, attr->unit, &rxru, &rxrp);

			tx_rate = delta_tx / ts_to_float(&delta_time);
			txr = cancel_down(tx_rate, attr->unit, &txru, &txrp);
		}

		rx = cancel_down(attr->rx_cnt, attr->unit, &rxu, &rxp);
		tx = cancel_down(attr->tx_cnt, attr->unit, &txu, &txp);

		printf("    %s %s", attr->name, attr->hts);

		if (!no_deltas) {
			char ds[128];
			printf(" (+%s)\n",
			       timestamp2str(&delta_time, ds, sizeof(ds)));
		} else
			printf("\n");

		printf("      RX Total: %.*f%-s", rxp, rx, rxu);

		if (!no_deltas)
			printf(" (+%.*f%-2s) Rate: %.*f%-2s/s\n",
			       drxp, drx, drxu, rxrp, rxr, rxru);
		else
			printf("\n");

		printf("      TX Total: %.*f%-3s", txp, tx, txu);

		if (!no_deltas)
			printf(" (+%.*f%-2s) Rate: %.*f%-2s/s\n",
			       dtxp, dtx, dtxu, txrp, txr, txru);
		else
			printf("\n");

		last_attr = attr->name;
		last_rx = attr->rx_cnt;
		last_tx = attr->tx_cnt;
		memcpy(&last_date, &cur_date, sizeof(last_date));
	}

	bdb_destroy_collection(&attrs);

	return 0;
}

static int print_attrs(struct db_item_full *item, struct db_filter *filter,
		       struct db_hist_params *params,
		       struct db_token *attr_filter)
{
	struct db_collection attrs;
	struct db_attr_full *attr;
	int rxp, txp;
	char *rxu, *txu;
	double rx, tx;

	if (in_history_mode(filter, params))
		return print_attr_hist(item, filter, params, attr_filter);

	if (bdb_fetch_attrs(db_conn, item->db_id, attr_filter, &attrs) < 0)
		return -1;

	for_each_entry(attr, &attrs, struct db_attr_full) {
		rx = cancel_down(attr->rx_rate, attr->unit, &rxu, &rxp);
		tx = cancel_down(attr->tx_rate, attr->unit, &txu, &txp);
		printf("            %-12s %8.*f%-3s %8.*f%-3s",
		       attr->txt, rxp, rx, rxu, txp, tx, txu);

		rx = cancel_down(attr->rx_cnt, attr->unit, &rxu, &rxp);
		tx = cancel_down(attr->tx_cnt, attr->unit, &txu, &txp);
		printf(" %8.*f%-3s %8.*f%-3s\n",
		       rxp, rx, rxu, txp, tx, txu);
	}

	bdb_destroy_collection(&attrs);

	return 0;
}

static void print_item(struct db_item_full *item, struct db_collection *items,
		       struct db_filter *filter, struct db_hist_params *params)

{
	char name[64];
	struct db_item_full *p;

	printf("  %-30s\n",
	       prepare_itemname(item, name, sizeof(name)));

	if (filter->f_attr) {
		if (print_attrs(item, filter, params, filter->f_attr) < 0)
			goto skip;
	} else {
		struct db_token bytes = { .t_name = "bytes" };
		struct db_token packets = { .t_name = "packets" };

		if (print_attrs(item, filter, params, &bytes) < 0)
			goto skip;

		if (print_attrs(item, filter, params, &packets) < 0)
			goto skip;
	}

skip:
	for_each_entry(p, items, struct db_item_full)
		if (p->parent == item->db_id)
			print_item(p, items, filter, params);
}

static int print_items(struct db_filter *filter, struct db_collection *items,
		       struct db_hist_params *params)
{
	struct db_item_full *item;

	for_each_entry(item, items, struct db_item_full) {
		/* skip over childs, we dump them recursively */
		if (item->parent >= 0)
			continue;

		print_item(item, items, filter, params);
	}

	return 0;
}


static int do_get_attr(struct db_filter *filter, struct db_hist_params *params)
{
	struct db_collection nodes;
	struct db_node_full *node;

	if (!in_history_mode(filter, params))
		printf("                                   Rate/s   " \
		       "                Total\n" \
		       "Name        Attribute          RX          TX" \
		       "          RX          TX\n" \
		       "---------------------------------------------" \
		       "---------------------------\n");
			    

	if (bdb_fetch_nodes(db_conn, filter->f_node, &nodes) < 0)
		return -1;

	for_each_entry(node, &nodes, struct db_node_full) {
		char name[64];
		struct db_collection items;

		printf("%s\n",
		    prepare_nodename(node, name, sizeof(name)));

		if (bdb_fetch_items(db_conn, node->db_id, filter->f_item, &items) < 0)
			continue;

		if (print_items(filter, &items, params) < 0)
			continue;

		bdb_destroy_collection(&items);
	}

	bdb_destroy_collection(&nodes);
	
	return 0;
}

static int do_get(int argc, char *argv[])
{
	int err;
	struct db_hist_params params;
	struct db_hist_params *pp;

	memset(&params, 0, sizeof(params));

	for (;;) {
		const char *gostr = "f:t:";
#ifdef HAVE_GETOPT_LONG
		struct option long_opts[] = {
			{"from", 1, 0, 'f'},
			{"to", 1, 0, 't'},
			{0, 0, 0, 0},
		};
		int c = getopt_long(argc, argv, gostr, long_opts, NULL);
#else
		int c = getopt(argc, argv, gostr);
#endif
		if (c == -1)
			break;

		switch (c) {
			case 'f':
				if (bdb_parse_date(optarg, &params.hp_from) < 0)
					return -1;
				params.hp_flags |= HP_HAS_FROM;
				break;

			case 't':
				if (bdb_parse_date(optarg, &params.hp_to) < 0)
					return -1;
				params.hp_flags |= HP_HAS_TO;
				break;

			default:
				return -1;
		}
	}
	
	while (optind < argc) {
		struct db_filter *f = bdb_parse_filter(argv[optind++]);

		if (f == NULL)
			return -1;

		pp = has_params(&params) ? &params : NULL;

		if (f->f_item)
			err = do_get_attr(f, pp);
		else
			err = do_get_node(f, pp);

		if (err < 0)
			return err;
	}

	return 0;
}

static int do_reinstall_rules(void)
{
	struct db_rule_settings s = {
		.up_read = 0,
		.up_sec = 0,
		.up_min = 0,
		.up_hour = 0,
		.up_day = 0,
		.expire_read = "60 seconds",
		.expire_sec = "5 minutes",
		.expire_min = "1 hour",
		.expire_hour = "1 month",
		.expire_day = "never",
	};
	
	return bdb_reinstall_rules(db_conn, &s);
}

static void print_usage(void)
{
	print_version();
	printf("\n%s", usage_text);
	exit(1);
}


int main(int argc, char *argv[])
{
	for (;;)
	{
		const char *gostr = "+hvVH:u:n:p:";

#ifdef HAVE_GETOPT_LONG
		struct option long_opts[] = {
			{"help", 0, 0, 'h'},
			{"version", 0, 0, 'v'},
			{"host", 1, 0, 'H'},
			{"user", 1, 0, 'u'},
			{"dbname", 1, 0, 'n'},
			{"password", 1, 0, 'p'},
			{0, 0, 0, 0},
		};
		int c = getopt_long(argc, argv, gostr, long_opts, NULL);
#else
		int c = getopt(argc, argv, gostr;
#endif

		if (c == -1)
			break;
		
		switch (c)
		{
			case 'H':
				set_db_host(optarg);
				break;

			case 'u':
				set_db_user(optarg);
				break;

			case 'n':
				set_db_name(optarg);
				break;

			case 'p':
				set_db_password(optarg);
				break;

			case 'f':
				set_configfile(optarg);
				break;

			case 'h':
				print_version();
				printf("\n%s", usage_text);
				exit(0);

			case 'V':
			case 'v':
				print_version();
				exit(0);

			default:
				exit(1);
		}
	}

	read_configfile();

	if (optind >= argc)
		print_usage();

	db_conn = bdb_connect(get_db_host(), get_db_user(), get_db_password(),
			get_db_name());
	if (db_conn == NULL)
		return -1;

	if (!strcasecmp(argv[optind], "create")) {
		optind++;
		if (do_create_tables() < 0)
			return -1;
	} else if (!strcasecmp(argv[optind], "destroy")) {
		optind++;
		if (do_destroy() < 0)
			return -1;
	} else if (!strcasecmp(argv[optind], "reinstall_rules")) {
		if (do_reinstall_rules() < 0)
			return -1;
	} else if (!strcasecmp(argv[optind], "dump")) {
		optind++;
		if (do_dump(argc, argv) < 0)
			return -1;
	} else if (!strcasecmp(argv[optind], "get")) {
		optind++;
		if (do_get(argc, argv) < 0)
			return -1;
	} else if (!strcasecmp(argv[optind], "analyze")) {
		optind++;
		if (do_analyze() < 0)
			return -1;
	} else
		print_usage();

	bdb_shutdown(db_conn);

	return 0;
}
