/*
 * db.c			Database Layer
 *
 * Copyright (c) 2001-2005 Thomas Graf <tgraf@suug.ch>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <bmon/bmon.h>
#include <bmon/db.h>
#include <bmon/node.h>
#include <bmon/output.h>
#include <bmon/graph.h>
#include <bmon/attr.h>
#include <bmon/input.h>
#include <bmon/utils.h>

static const char *hist_tables[] = {
	"hist_r",
	"hist_s",
	"hist_m",
	"hist_h",
	"hist_d",
	NULL
};

/* output buffer of yacc db filter parsing */
struct db_token *db_filter_out;

static struct db_node *node_cache;
static int valid_node_cache;
static int node_cache_size;

static struct db_item *item_cache;
static int valid_item_cache;
static int item_cache_size;

static struct db_attr *attr_cache;
static int valid_attr_cache;
static int attr_cache_size;

int foreach_table(int (*cb)(PGconn *, const char *, void *), PGconn *conn,
		  void *arg)
{
	int err;
	const char **p;

	for (p = hist_tables; *p; p++) {
		err = cb(conn, *p, arg);
		if (err < 0)
			return err;
	}

	return 0;
}

static inline PGresult * xPQexecParams(PGconn *conn, int nparams, char **params,
				       char *stmt, ...)
{
	va_list args;
	char buf[1024];

	va_start(args, stmt);
	vsnprintf(buf, sizeof(buf), stmt, args);
	va_end(args);

	return PQexecParams(conn, buf, nparams, NULL,
        		    (const char * const *) params, NULL, NULL, 0);
}

static inline int db_get_int(PGresult *res, int row, int col)
{
	char *data = PQgetvalue(res, row, col);
	return strtol(data, NULL, 0);
}

static inline uint64_t db_get_bigint(PGresult *res, int row, int col)
{
	char *data = PQgetvalue(res, row, col);
	return strtoll(data, NULL, 0);
}

static inline char * db_get_charp(PGresult *res, int row, int col)
{
	return PQgetvalue(res, row, col);
}

static char * db_get_timestamp(PGresult *res, int row, int col)
{
	return PQgetvalue(res, row, col);
}

static int update_node_cache(PGconn *conn)
{
	int err = -1, i;
	int i_id, i_name;
	PGresult *r;

	valid_node_cache = 0;
	r = PQexec(conn, "SELECT id, name FROM nodes ORDER BY name");

	if (PQresultStatus(r) != PGRES_TUPLES_OK) {
		xwarn("Could not update node cache: %s",
		      PQerrorMessage(conn));
		goto errout;
	}

	node_cache_size = PQntuples(r);
	node_cache = realloc(node_cache,
			     PQntuples(r) * sizeof(struct db_node));
	if (node_cache == NULL) {
		err = -ENOMEM;
		goto errout;
	}

	i_id = PQfnumber(r, "id");
	i_name = PQfnumber(r, "name");

	for (i = 0; i < PQntuples(r); i++) {
		int id = db_get_int(r, i, i_id);
		char *name = db_get_charp(r, i, i_name);

		node_cache[i].index = i;
		node_cache[i].db_id = id;
		strncpy(node_cache[i].name, name, DB_NODELEN-1);
        }

	valid_node_cache = 1;
	err = 0;

errout:
	PQclear(r);
	return err;
}

static int update_item_cache(PGconn *conn)
{
        int err = -1, i, i_id, i_handle, i_parent, i_node, i_name;
        PGresult *r;

	valid_item_cache = 0;
        r = PQexec(conn, "SELECT id, name, handle, parent, node FROM items ORDER BY id");

	if (PQresultStatus(r) != PGRES_TUPLES_OK) {
		xwarn("Could not update item cache: %s", PQerrorMessage(conn));
		goto errout;
	}

        item_cache_size = PQntuples(r);
        item_cache = realloc(item_cache,
                             PQntuples(r) * sizeof(struct db_item));
        if (item_cache == NULL) {
                err = -ENOMEM;
                goto errout;
        }
        
        i_id = PQfnumber(r, "id");
        i_handle = PQfnumber(r, "handle");
        i_parent = PQfnumber(r, "parent");
        i_node = PQfnumber(r, "node");
        i_name = PQfnumber(r, "name");

        for (i = 0; i < PQntuples(r); i++) {
                int parent;
                int id = db_get_int(r, i, i_id);
                int handle = db_get_int(r, i, i_handle);
                int node = db_get_int(r, i, i_node);
                char *name = db_get_charp(r, i, i_name);

		if (PQgetisnull(r, i, i_parent))
			parent = -1;
		else
			parent = db_get_int(r, i, i_parent);

                item_cache[i].index = i;
                item_cache[i].db_id = id;
                item_cache[i].handle = handle;
                item_cache[i].parent = parent;
                item_cache[i].node = node;
                strncpy(item_cache[i].name, name, DB_ITEMLEN-1);
        }

	valid_item_cache = 1;
        err = 0;
errout:
        PQclear(r);

	return err;
}

struct db_cnt * bdb_get_counters(PGconn *conn, int *len)
{
	char *params[1];
        PGresult *r;
	int ncnts;
	struct db_cnt *cnts;
	int i_name, i_handle, i_parent, i_attr_type, i_rx_cnt, i_tx_cnt;
	int i_id, i_last_update, i_indent, i_major, i_minor, i;

	params[0] = get_local_node()->n_name;

	valid_item_cache = 0;
        r = xPQexecParams(conn, 1, params,
			  "SELECT i.id, i.name AS item_name, i.handle, " \
			  "       i.parent, i.indent, i.major_attr, " \
			  "       i.minor_attr, a.name AS attr_type, " \
			  "       a.rx_cnt, a.tx_cnt, a.last_update " \
			  "FROM nodes AS n, items AS i, attrs AS a " \
			  "WHERE n.name = $1 AND n.id = i.node AND " \
			  "      i.id = a.item " \
			  "ORDER BY i.id, a.name");
			  
	if (PQresultStatus(r) != PGRES_TUPLES_OK) {
		xwarn("Could not fetch counters: %s", PQerrorMessage(conn));
		return NULL;
	}

	ncnts = PQntuples(r);
	cnts = xcalloc(ncnts, sizeof(struct db_cnt));

	i_id = PQfnumber(r, "id");
	i_name = PQfnumber(r, "item_name");
        i_handle = PQfnumber(r, "handle");
        i_parent = PQfnumber(r, "parent");
	i_indent = PQfnumber(r, "indent");
	i_major = PQfnumber(r, "major_attr");
	i_minor = PQfnumber(r, "minor_attr");
	i_attr_type = PQfnumber(r, "attr_type");
	i_rx_cnt = PQfnumber(r, "rx_cnt");
	i_tx_cnt = PQfnumber(r, "tx_cnt");
	i_last_update = PQfnumber(r, "last_update");

        for (i = 0; i < ncnts; i++) {
		cnts[i].item_index = db_get_int(r, i, i_id);
		cnts[i].item_handle = db_get_int(r, i, i_handle);
		cnts[i].item_indent = db_get_int(r, i, i_indent);
		cnts[i].item_major_attr = db_get_int(r, i, i_major);
		cnts[i].item_minor_attr = db_get_int(r, i, i_minor);
		cnts[i].attr_rx_cnt = db_get_bigint(r, i, i_rx_cnt);
		cnts[i].attr_tx_cnt = db_get_bigint(r, i, i_tx_cnt);
		strncpy(cnts[i].item_name, db_get_charp(r, i, i_name),
			DB_ITEMLEN-1);
		strncpy(cnts[i].attr_type, db_get_charp(r, i, i_attr_type),
			DB_ATTRLEN-1);
		strncpy(cnts[i].last_update,
			db_get_timestamp(r, i, i_last_update), DB_TSLEN-1);

		if (PQgetisnull(r, i, i_parent))
			cnts[i].item_parent = -1;
		else
			cnts[i].item_parent = db_get_int(r, i, i_parent);
        }

        PQclear(r);
	*len = ncnts;

	return cnts;
}

static int update_attr_cache(PGconn *conn)
{
        int err = -1, i, i_id, i_item, i_name, i_node;
        PGresult *r;

	valid_attr_cache = 0;
        r = PQexec(conn, "SELECT a.id, a.item, a.name, i.node " \
			    "FROM attrs AS a, items AS i " \
			    "WHERE a.item = i.id ORDER BY a.name");

	if (PQresultStatus(r) != PGRES_TUPLES_OK) {
		xwarn("Could not update attr cache: %s", PQerrorMessage(conn));
		goto errout;
	}

        attr_cache_size = PQntuples(r);
        attr_cache = realloc(attr_cache,
                             PQntuples(r) * sizeof(struct db_attr));
        if (attr_cache == NULL) {
                err = -ENOMEM;
                goto errout;
        }
        
        i_id = PQfnumber(r, "id");
        i_item = PQfnumber(r, "item");
        i_name = PQfnumber(r, "name");
        i_node = PQfnumber(r, "node");

        for (i = 0; i < PQntuples(r); i++) {
                int id = db_get_int(r, i, i_id);
                int item = db_get_int(r, i, i_item);
                char *name = db_get_charp(r, i, i_name);

                attr_cache[i].index = i;
                attr_cache[i].db_id = id;
                attr_cache[i].item = item;
                strncpy(attr_cache[i].name, name, DB_ATTRLEN-1);
        }

	valid_attr_cache = 1;
        err = 0;
errout:
        PQclear(r);

	return err;
}

int bdb_update_caches(PGconn *conn)
{
	if (update_node_cache(conn) < 0 ||
	    update_item_cache(conn) < 0 ||
            update_attr_cache(conn) < 0)
		return -1;

	return 0;
}

struct db_node * bdb_get_node_cache(void)
{
	if (!valid_node_cache)
		BUG();

	return node_cache;
}

int bdb_get_node_cache_rows(void)
{
	if (!valid_node_cache)
		BUG();

	return node_cache_size;
}

struct db_item * bdb_get_item_cache(void)
{
	if (!valid_item_cache)
		BUG();

	return item_cache;
}

int bdb_get_item_cache_rows(void)
{
	if (!valid_item_cache)
		BUG();

	return item_cache_size;
}

struct db_attr * bdb_get_attr_cache(void)
{
	if (!valid_attr_cache)
		BUG();

	return attr_cache;
}

int bdb_get_attr_cache_rows(void)
{
	if (!valid_attr_cache)
		BUG();

	return attr_cache_size;
}

int bdb_get_node(const char *name)
{
	int i;

	if (!valid_node_cache)
		BUG();

	for (i = 0; i < node_cache_size; i++)
		if (!strcmp(node_cache[i].name, name))
			return node_cache[i].db_id;

	return -1;
}

int bdb_find_node(struct node *node)
{
	return bdb_get_node(node->n_name);
}

int bdb_get_item(int node_id, const char *name, uint32_t handle, int parent)
{
	int i;

	if (!valid_item_cache)
		BUG();

	for (i = 0; i < item_cache_size; i++)
		if (item_cache[i].node == node_id &&
		    item_cache[i].handle == handle &&
		    (parent < 0 || parent == item_cache[i].parent) &&
		    !strcmp(item_cache[i].name, name))
			return item_cache[i].db_id;

	return -1;
}

int bdb_find_item(struct item *it, struct node *node, int node_id)
{
	int parent = -1;

	if (it->i_flags & ITEM_FLAG_IS_CHILD) {
		struct item *p = get_item(node, it->i_parent);
		if (p)
			parent = bdb_find_item(p, node, node_id);
	}

	return bdb_get_item(node_id, it->i_name, it->i_handle, parent);
}

int bdb_get_attr(const char *name, int item_id)
{
	int i;

	if (!valid_attr_cache)
		BUG();

	for (i = 0; i < attr_cache_size; i++)
		if (attr_cache[i].item == item_id &&
		    !strcmp(attr_cache[i].name, name))
			return attr_cache[i].db_id;

	return -1;
}

int bdb_find_attr(struct stat_attr *attr, int item_id)
{
	return bdb_get_attr(attr_type2name(attr->a_type), item_id);
}

int bdb_insert_node(PGconn *conn, struct node *n)
{
	char *params[] = {n->n_name, n->n_from};

	PGresult *r = xPQexecParams(conn, 2, params, 
        			   "INSERT INTO nodes (name, source, created, last_update) " \
                                   "VALUES ($1, $2, NOW(), NOW())");

	if (PQresultStatus(r) != PGRES_COMMAND_OK) {
		xwarn("Could not insert node \"%s\": %s", n->n_name, PQerrorMessage(conn));
		goto errout;
	}

        PQclear(r);

	if (update_node_cache(conn) < 0)
		return -1;

        return bdb_find_node(n);

errout:
        PQclear(r);
        return -1;
}

int bdb_insert_item(PGconn *conn, struct item *it, struct node *node,
		    int node_id)
{
	int parent = -1;
        char sparent[16];
	char *params[] = {it->i_name, NULL, it->i_desc};

	if (it->i_flags & ITEM_FLAG_IS_CHILD) {
		struct item *p = get_item(node, it->i_parent);
		if (p) {
			parent = bdb_find_item(p, node, node_id);
                        snprintf(sparent, sizeof(sparent), "%d", parent);
                        params[1] = sparent;
                }
	}

	PGresult *r = xPQexecParams(conn, 3, params,
        			   "INSERT INTO items (name, node, handle, parent, " \
                                   "indent, description, created, last_update) " \
                                   "VALUES ($1, %d, %u, $2, %d, $3, NOW(), NOW())",
                                   node_id, it->i_handle, it->i_level);

	if (PQresultStatus(r) != PGRES_COMMAND_OK) {
		xwarn("Could not insert item \"%s\": %s", it->i_name, PQerrorMessage(conn));
		goto errout;
	}

        PQclear(r);

	if (update_item_cache(conn) < 0)
		return -1;

	return bdb_find_item(it, node, node_id);

errout:
        PQclear(r);
        return -1;
}

int bdb_update_item(PGconn *conn, struct item *it, int item_id)
{
        int err = 0;
	PGresult *r;
	
	r = xPQexecParams(conn, 0, NULL,
			  "UPDATE items " \
			  "SET rx_usage = %d, tx_usage = %d, " \
			  "    major_attr = %d, minor_attr = %d, " \
			  "    last_update = NOW() " \
			  "WHERE id = %d",
			  it->i_rx_usage, it->i_tx_usage,
			  it->i_major_attr, it->i_minor_attr, item_id);

	if (PQresultStatus(r) != PGRES_COMMAND_OK) {
		xwarn("Could not update item \"%s\": %s",
		      it->i_name, PQerrorMessage(conn));
                err = -1;
	}

        PQclear(r);

        return err;

}

int bdb_finish_updates(PGconn *conn)
{
	int err = 0;

	PGresult *r = PQexec(conn, "UPDATE writes SET ts = NOW(), " \
			     "needs_update = 0 WHERE needs_update = 1");

	if (PQresultStatus(r) != PGRES_COMMAND_OK) {
		xwarn("Could not finalize updates:", PQerrorMessage(conn));
		err = -1;
	}

	PQclear(r);

	return err;
}

int bdb_insert_attr(PGconn *conn, struct stat_attr *a, int item_id)
{
        char *params[1];
        PGresult *r;

        params[0] = (char *) attr_type2name(a->a_type);

        r = xPQexecParams(conn, 1, params,
        		  "INSERT INTO attrs (name, item, rx_rate, tx_rate, " \
                          "rx_cnt, tx_cnt, created, last_update) " \
                          "VALUES ($1, %d, %d, %d, %" PRIu64 ", %" PRIu64 ", NOW(), " \
                          "NOW())", item_id, attr_get_rx_rate(a), attr_get_tx_rate(a),
                          attr_get_rx(a), attr_get_tx(a));

	if (PQresultStatus(r) != PGRES_COMMAND_OK) {
		xwarn("Could not insert attr \"%s:%d\": %s",
		       params[0], item_id, PQerrorMessage(conn));
		goto errout;
        }

        PQclear(r);

	if (update_attr_cache(conn) < 0)
		return -1;

	return bdb_find_attr(a, item_id);

errout:
        PQclear(r);
        return -1;
}

int bdb_update_attr(PGconn *conn, struct stat_attr *a, int attr_id)
{
	int err = 0;

	PGresult *r = xPQexecParams(conn, 0, NULL,
				   "UPDATE attrs SET rx_rate = %d, tx_rate = %d, " \
                                   "rx_cnt = %" PRIu64 ", tx_cnt = %" PRIu64 ", " \
                                   "last_update = NOW() WHERE id = %d",
                                   attr_get_rx_rate(a), attr_get_tx_rate(a),
                                   attr_get_rx(a), attr_get_tx(a),  attr_id);

	if (PQresultStatus(r) != PGRES_COMMAND_OK) {
		xwarn("Could not update attr %d: %s", attr_id,
		      PQerrorMessage(conn));
                err = -1;
        }

        PQclear(r);

        return err;
}

static int create_hist_table(PGconn *conn, const char *name, void *arg)
{
	int err = 0;
	PGresult *r;

	r = xPQexecParams(conn, 0, NULL,
        	"CREATE TABLE %s (" \
                "       attr            integer         NOT NULL," \
                "       ts              timestamp(6)    NOT NULL," \
                "       rx_cnt          int8            NOT NULL," \
                "       tx_cnt          int8            NOT NULL," \
                "       permanent       int2            DEFAULT 0," \
                "       PRIMARY KEY (attr, ts)," \
                "       FOREIGN KEY (attr) REFERENCES attrs(id)" \
                ")", name);

	if (PQresultStatus(r) != PGRES_COMMAND_OK)
		err = -1;

	PQclear(r);

	return err;
}

int bdb_create_tables(PGconn *conn)
{
	int err = -1;
	PGresult *r;

#define PROCESS_RESULT(r) \
	if (PQresultStatus(r) != PGRES_COMMAND_OK) goto errout; else PQclear(r)

	r = PQexec(conn,
        	"CREATE TABLE nodes (" \
                "       id              serial," \
                "       name            text            NOT NULL," \
                "       source          text," \
                "       created         timestamp(6)    NOT NULL," \
                "       last_update     timestamp(6)    NOT NULL," \
                "       PRIMARY KEY (id)" \
                ")");
        PROCESS_RESULT(r);
	
	r = PQexec(conn,
        	"CREATE TABLE items (" \
                "       id              serial," \
                "       name            text            NOT NULL," \
                "       description     text," \
                "       node            integer         NOT NULL," \
                "       handle          int4," \
                "       parent          integer," \
                "       indent          int4            DEFAULT 0," \
                "       rx_usage        int2," \
                "       tx_usage        int2," \
		"       major_attr      int4," \
		"       minor_attr      int4," \
                "       created         timestamp(6)    NOT NULL," \
                "       last_update     timestamp(6)    NOT NULL," \
                "       PRIMARY KEY (id)," \
                "       FOREIGN KEY (node) REFERENCES nodes(id)," \
                "       FOREIGN KEY (parent) REFERENCES items(id)" \
                ")");
        PROCESS_RESULT(r);

	r = PQexec(conn,
        	"CREATE TABLE attrs (" \
                "       id              bigserial," \
                "       item            integer         NOT NULL," \
                "       name            varchar(15)     NOT NULL," \
                "       rx_rate         int4," \
                "       tx_rate         int4," \
                "       rx_cnt          int8            NOT NULL," \
                "       tx_cnt          int8            NOT NULL," \
                "       created         timestamp(6)    NOT NULL," \
                "       last_update     timestamp(6)    NOT NULL," \
                "       PRIMARY KEY (id)," \
                "       FOREIGN KEY (item) REFERENCES items(id)" \
                ")");
        PROCESS_RESULT(r);

	r = PQexec(conn,
        	"CREATE TABLE attr_desc (" \
                "       id              varchar(15)," \
                "       unit            int2            NOT NULL," \
                "       txt             text            NOT NULL," \
                "       PRIMARY KEY(id)" \
                ")");
        PROCESS_RESULT(r);

	r = PQexec(conn,
        	"CREATE TABLE writes (" \
                "       id              int," \
		"       ts              timestamp       NOT NULL," \
		"       needs_update    int             DEFAULT 0," \
                "       PRIMARY KEY(id)" \
                ")");
        PROCESS_RESULT(r);

	r = PQexec(conn,
		   "INSERT INTO writes (id, ts) VALUES (115, '1970-01-01')");
        PROCESS_RESULT(r);
	r = PQexec(conn,
		   "INSERT INTO writes (id, ts) VALUES (109, '1970-01-01')");
        PROCESS_RESULT(r);
	r = PQexec(conn,
		   "INSERT INTO writes (id, ts) VALUES (104, '1970-01-01')");
        PROCESS_RESULT(r);
	r = PQexec(conn,
		   "INSERT INTO writes (id, ts) VALUES (100, '1970-01-01')");
        PROCESS_RESULT(r);

	r = NULL;
	if (foreach_table(&create_hist_table, conn, NULL) < 0)
		goto errout;

	err = 0;
errout:
	if (r)
		PQclear(r);

	if (err) {
		xwarn("Could not create table: %s", PQerrorMessage(conn));
		xwarn("Dropping tables....\n");
		return bdb_drop_tables(conn);
	}

	return 0;
#undef PROCESS_RESULT
}

static int drop_table(PGconn *conn, const char *name, void *arg)
{
	PGresult *r;

	r = xPQexecParams(conn, 0, NULL, "DROP TABLE %s", name);

	if (PQresultStatus(r) != PGRES_COMMAND_OK)
		xwarn("Could not drop table \"%s\": %s",
		      name, PQerrorMessage(conn));

	PQclear(r);

	return 0;
}

int bdb_drop_tables(PGconn *conn)
{
	PGresult *r;

#define PROCESS_RESULT(r)  \
	if (PQresultStatus(r) != PGRES_COMMAND_OK) \
		xwarn("Could not drop table: %s", PQerrorMessage(conn)); \
	PQclear(r)

	r = PQexec(conn, "DROP TABLE writes");
        PROCESS_RESULT(r);

	foreach_table(&drop_table, conn, NULL);

	r = PQexec(conn, "DROP TABLE attr_desc");
        PROCESS_RESULT(r);

	r = PQexec(conn, "DROP TABLE attrs");
        PROCESS_RESULT(r);

	r = PQexec(conn, "DROP TABLE items");
        PROCESS_RESULT(r);

	r = PQexec(conn, "DROP TABLE nodes");
        PROCESS_RESULT(r);

#undef PROCESS_RESULT

	/* We always succeed but might have warnings */
	return 0;
}

int create_expire_rule(PGconn *conn, const char *name, int do_rule,
		       const char *expire)
{
	int err = 0;
	PGresult *r;

	if (expire && strcasecmp(expire, "never") && do_rule) {
		char *e = malloc((strlen(expire)*2)+1);

		PQescapeString(e, expire, strlen(expire));

		r = xPQexecParams(conn, 0, NULL,
			   "CREATE OR REPLACE RULE expire_%s AS " \
			   "ON INSERT TO %s " \
			   "DO DELETE FROM %s WHERE permanent = 0 AND " \
			   "ts < (NOW() - INTERVAL '%s')",
			   name, name, name, e);

		xfree(e);

		if (PQresultStatus(r) != PGRES_COMMAND_OK) {
			xwarn("Could not create rule: %s",
			      PQerrorMessage(conn)); \
			err = -1;
		}
	} else
		r = xPQexecParams(conn, 0, NULL, "DROP RULE expire_%s ON %s",
				  name, name);

	PQclear(r);

	return err;
}

static int create_update_rule(PGconn *conn, const char *name, int do_rule,
			      const char *delta, int code)
{
	int err = 0;
	PGresult *r;

	if (do_rule) {
		r = xPQexecParams(conn, 0, NULL,
			   "CREATE OR REPLACE RULE up_%s AS " \
			   "ON UPDATE TO attrs " \
			   "WHERE (new.last_update - INTERVAL '%s') >= " \
			   "   (SELECT ts FROM writes WHERE id = %d) " \
			   "DO (INSERT INTO %s (attr, ts, rx_cnt, tx_cnt)" \
			   " VALUES (new.id, new.last_update, new.rx_cnt, " \
			   "new.tx_cnt); UPDATE writes SET needs_update = 1 " \
			   "WHERE id = %d)",
			   name, delta, code, name, code);

		if (PQresultStatus(r) != PGRES_COMMAND_OK) {
			xwarn("Could not create rule: %s",
			      PQerrorMessage(conn));
			err = -1;
		}
	} else
		r = xPQexecParams(conn, 0, NULL, "DROP RULE up_%s ON attrs",
				  name);

	PQclear(r);

	return err;
}

int bdb_reinstall_rules(PGconn *conn, struct db_rule_settings *s)
{
	PGresult *r;

	create_expire_rule(conn, "hist_r", s->up_read, s->expire_read);
	create_expire_rule(conn, "hist_s", s->up_sec, s->expire_sec);
	create_expire_rule(conn, "hist_m", s->up_min, s->expire_min);
	create_expire_rule(conn, "hist_h", s->up_hour, s->expire_hour);
	create_expire_rule(conn, "hist_d", s->up_day, s->expire_day);

	if (s->up_read) {
		r = PQexec(conn,
			   "CREATE OR REPLACE RULE up_hist_r AS " \
			   "ON UPDATE TO attrs " \
			   "DO INSERT INTO hist_r (attr, ts, rx_cnt, tx_cnt) " \
			   "VALUES (new.id, new.last_update, new.rx_cnt, " \
			   "new.tx_cnt)");
		if (PQresultStatus(r) != PGRES_COMMAND_OK)
			xwarn("Could not create rule: %s",
			      PQerrorMessage(conn));

		PQclear(r);
	} else {
		r = PQexec(conn, "DROP RULE up_hist_r ON attrs");
		PQclear(r);
	}

	create_update_rule(conn, "hist_s", s->up_sec, "1 second", 115);
	create_update_rule(conn, "hist_m", s->up_min, "1 minute", 109);
	create_update_rule(conn, "hist_h", s->up_hour, "1 hour", 104);
	create_update_rule(conn, "hist_d", s->up_day, "1 day", 100);

	return 0;
}

static int drop_rule(PGconn *conn, const char *name, void *arg)
{
	PGresult *r;

	r = xPQexecParams(conn, 0, NULL, "DROP RULE up_%s ON attrs", name);

	if (PQresultStatus(r) != PGRES_COMMAND_OK)
		xwarn("Could not drop rule: %s", PQerrorMessage(conn));

	PQclear(r);

	return 0;
}

int bdb_drop_rules(PGconn *conn)
{
	return foreach_table(&drop_rule, conn, NULL);
}


static int insert_attr_desc(struct attr_type *a, void *arg)
{
	int err = -1;
	char *params[] = {a->at_name, a->at_desc};
	PGconn *conn = arg;

	PGresult *r = xPQexecParams(conn, 1, params,
				   "SELECT id FROM attr_desc WHERE id = $1");

	if (PQresultStatus(r) != PGRES_TUPLES_OK) {
		xwarn("Could not insert attr description for \"%s\": %s",
		      a->at_name, PQerrorMessage(conn));
		goto errout;
	}

	if (PQntuples(r) == 1) {
		err = 0;
		goto errout;
	}

	PQclear(r);

	r = xPQexecParams(conn, 2, params,
			 "INSERT INTO attr_desc (id, unit, txt) " \
			 "VALUES ($1, %d, $2)", a->at_unit);

	if (PQresultStatus(r) != PGRES_COMMAND_OK) {
		xwarn("Could not insert attr description for \"%s\": %s",
		      a->at_name, PQerrorMessage(conn));
		goto errout;
	}

	err = 0;
errout:
	PQclear(r);
	return err;
}

int bdb_insert_attr_descs(PGconn *conn)
{
	return foreach_attr_type(&insert_attr_desc, conn);
}

PGconn * bdb_connect(const char *host, const char *user, const char *password,
		     const char *dbname)
{
	char conninfo[256];
	PGconn *c;

	snprintf(conninfo, sizeof(conninfo),
		 "host='%s' dbname='%s' user='%s' password='%s'",
		 host, dbname, user, password);
	
	c = PQconnectdb(conninfo);

	if (PQstatus(c) != CONNECTION_OK)
		quit("Connection to database failed: %s",
			PQerrorMessage(c));

	return c;
}

void bdb_shutdown(PGconn *conn)
{
	PQfinish(conn);
}

int bdb_fetch_nodes(PGconn *conn, struct db_token *tok,
		    struct db_collection *col)
{
	int i, err = -1;
	struct db_node_full *nodes = NULL;
	int i_id, i_name, i_source, i_created, i_last_update;
	PGresult *r;

	if (tok && !(tok->t_flags & DB_T_WILDCARD)) {
		if (tok->t_name) {
			char *params[1] = {tok->t_name};
			r = xPQexecParams(conn, 1, params,
					  "SELECT id, name, source, " \
					  "created, last_update FROM nodes " \
					  "WHERE name = $1 ORDER BY name");
		} else
			r = xPQexecParams(conn, 0, NULL,
					  "SELECT id, name, source, " \
					  "created, last_update FROM nodes " \
					  "WHERE id = %d ORDER BY name", tok->t_id);
	} else
		r = xPQexecParams(conn, 0, NULL, "SELECT id, name, source, " \
				    "created, last_update FROM nodes ORDER BY name");
		
	if (PQresultStatus(r) != PGRES_TUPLES_OK) {
		xwarn("Could not fetch nodes: %s", PQerrorMessage(conn));
		goto errout;
	}

	nodes = xcalloc(PQntuples(r), sizeof(struct db_node_full));
	if (nodes == NULL)
		goto errout;

	col->size = PQntuples(r);
	col->array = nodes;

        i_id = PQfnumber(r, "id");
        i_name = PQfnumber(r, "name");
	i_source = PQfnumber(r, "source");
	i_created = PQfnumber(r, "created");
	i_last_update = PQfnumber(r, "last_update");
	
	for (i = 0; i < PQntuples(r); i++) {
		nodes[i].index = i;
		nodes[i].db_id = db_get_int(r, i, i_id);
		strncpy(nodes[i].name, db_get_charp(r, i, i_name), DB_NODELEN-1);
		strncpy(nodes[i].source, db_get_charp(r, i, i_source), DB_SOURCELEN-1);
		strncpy(nodes[i].created, db_get_timestamp(r, i, i_created), DB_TSLEN-1);
		strncpy(nodes[i].last_update, db_get_timestamp(r, i, i_last_update), DB_TSLEN-1);
        }

        err = 0;

errout:
        PQclear(r);

	return err;
}

int bdb_fetch_items(PGconn *conn, int node_id, struct db_token *tok,
		    struct db_collection *col)

{
	int i, err = -1;
	struct db_item_full *items = NULL;
	int i_id, i_name, i_desc, i_created, i_last_update, i_handle;
	int i_parent, i_indent, i_rx_usage, i_tx_usage;
	PGresult *r;

	if (tok && !(tok->t_flags & DB_T_WILDCARD)) {
		if (tok->t_name) {
			char *params[1] = {tok->t_name};
			r = xPQexecParams(conn, 1, params,
					  "SELECT id, name, description, " \
					  "handle, parent, indent, rx_usage, " \
					  "tx_usage, created, last_update " \
					  "FROM items WHERE node = %d AND " \
					  "name = $1", node_id);
		} else
			r = xPQexecParams(conn, 0, NULL,
					  "SELECT id, name, description, " \
					  "handle, parent, indent, rx_usage, " \
					  "tx_usage, created, last_update " \
					  "FROM items WHERE node = %d AND " \
					  "id = %d", node_id, tok->t_id);
	} else
		r = xPQexecParams(conn, 0, NULL,
		 		  "SELECT id, name, description, " \
				  "handle, parent, indent, rx_usage, " \
				  "tx_usage, created, last_update " \
				  "FROM items WHERE node = %d", node_id);

	if (PQresultStatus(r) != PGRES_TUPLES_OK) {
		xwarn("Could not fetch items: %s", PQerrorMessage(conn));
		goto errout;
	}

	items = xcalloc(PQntuples(r), sizeof(struct db_item_full));
	if (items == NULL)
		goto errout;

	col->size = PQntuples(r);
	col->array = items;

	i_id = PQfnumber(r, "id");
	i_name = PQfnumber(r, "name");
	i_desc = PQfnumber(r, "description");
	i_handle = PQfnumber(r, "handle");
	i_parent = PQfnumber(r, "parent");
	i_indent = PQfnumber(r, "indent");
	i_rx_usage = PQfnumber(r, "rx_usage");
	i_tx_usage = PQfnumber(r, "tx_usage");
	i_created = PQfnumber(r, "created");
	i_last_update = PQfnumber(r, "last_update");
	
	for (i = 0; i < PQntuples(r); i++) {
		items[i].index = i;
		items[i].node = node_id;
		items[i].db_id = db_get_int(r, i, i_id);
		items[i].handle = db_get_int(r, i, i_handle);
		items[i].indent = db_get_int(r, i, i_indent);
		items[i].rx_usage = db_get_int(r, i, i_rx_usage);
		items[i].tx_usage = db_get_int(r, i, i_tx_usage);
		if (PQgetisnull(r, i, i_parent))
			items[i].parent = -1;
		else
			items[i].parent = db_get_int(r, i, i_parent);
		strncpy(items[i].name, db_get_charp(r, i, i_name), DB_ITEMLEN-1);
		strncpy(items[i].desc, db_get_charp(r, i, i_desc), DB_DESCLEN-1);
		strncpy(items[i].created, db_get_timestamp(r, i, i_created), DB_TSLEN-1);
		strncpy(items[i].last_update, db_get_timestamp(r, i, i_last_update), DB_TSLEN-1);
	}

	err = 0;

errout:
	PQclear(r);

	return err;
}

int bdb_fetch_attrs(PGconn *conn, int item_id, struct db_token *tok,
		    struct db_collection *col)
{
	int i, err = -1;
	struct db_attr_full *attrs = NULL;
	int i_id, i_name, i_created, i_last_update, i_rx_rate, i_tx_rate;
	int i_rx_cnt, i_tx_cnt, i_unit, i_txt;
	PGresult *r;

	if (tok && !(tok->t_flags & DB_T_WILDCARD)) {
		if (tok->t_name) {
			char *params[1] = {tok->t_name};
			r = xPQexecParams(conn, 1, params,
					  "SELECT a.id, a.name, a.rx_rate, " \
					  "a.tx_rate, a.rx_cnt, a.tx_cnt, " \
					  "a.created, a.last_update, d.unit, " \
					  "d.txt FROM attrs AS a, attr_desc " \
					  "AS d WHERE d.id = a.name AND " \
					  "a.item = %d AND " \
					  "a.name = $1 ORDER BY a.name", item_id);
		} else
			r = xPQexecParams(conn, 0, NULL,
					  "SELECT a.id, a.name, a.rx_rate, a.tx_rate, " \
					  "a.rx_cnt, a.tx_cnt, a.created, a.last_update, " \
					  "d.unit, d.txt " \
					  "FROM attrs AS a, attr_desc AS d " \
					  "WHERE d.id = a.name AND a.item = %d AND " \
					  "a.id = %d ORDER BY a.name", item_id, tok->t_id);
	} else
		r = xPQexecParams(conn, 0, NULL,
				  "SELECT a.id, a.name, a.rx_rate, a.tx_rate, " \
				  "a.rx_cnt, a.tx_cnt, a.created, a.last_update, " \
				  "d.unit, d.txt " \
				  "FROM attrs AS a, attr_desc AS d " \
				  "WHERE d.id = a.name AND a.item = %d " \
				  "ORDER BY a.name", item_id);

	if (PQresultStatus(r) != PGRES_TUPLES_OK) {
		xwarn("Could not fetch attributes: %s", PQerrorMessage(conn));
		goto errout;
	}

	attrs = xcalloc(PQntuples(r), sizeof(struct db_attr_full));
	if (attrs == NULL)
		goto errout;

	col->size = PQntuples(r);
	col->array = attrs;
	
	i_id = PQfnumber(r, "id");
	i_name = PQfnumber(r, "name");
	i_rx_rate = PQfnumber(r, "rx_rate");
	i_tx_rate = PQfnumber(r, "tx_rate");
	i_rx_cnt = PQfnumber(r, "rx_cnt");
	i_tx_cnt = PQfnumber(r, "tx_cnt");
	i_unit = PQfnumber(r, "unit");
	i_txt = PQfnumber(r, "txt");
	i_created = PQfnumber(r, "created");
	i_last_update = PQfnumber(r, "last_update");
	
	for (i = 0; i < PQntuples(r); i++) {
		attrs[i].index = i;
		attrs[i].item = item_id;
		attrs[i].db_id = db_get_int(r, i, i_id);
		attrs[i].unit = db_get_int(r, i, i_unit);
		attrs[i].rx_rate = db_get_int(r, i, i_rx_rate);
		attrs[i].tx_rate = db_get_int(r, i, i_tx_rate);
		attrs[i].rx_cnt = db_get_bigint(r, i, i_rx_cnt);
		attrs[i].tx_cnt = db_get_bigint(r, i, i_tx_cnt);
		strncpy(attrs[i].name, db_get_charp(r, i, i_name), DB_ATTRLEN-1);
		strncpy(attrs[i].created, db_get_timestamp(r, i, i_created), DB_TSLEN-1);
		strncpy(attrs[i].last_update, db_get_timestamp(r, i, i_last_update), DB_TSLEN-1);
		strncpy(attrs[i].txt, db_get_charp(r, i, i_txt), DB_TXTLEN-1);
	}
	
	err = 0;

errout:
	PQclear(r);

	return err;
}

static inline int valid_hist_table(const char *name)
{
	char t = tolower(name[0]);
	return t == 'r' || t == 's' || t == 'm' || t == 'h' || t == 'd';
}

int bdb_fetch_attr_hist(PGconn *conn, int item_id, struct db_token *tok,
			struct db_collection *col,
			struct db_hist_params *params, const char *table)
{
	int i, err = -1;
	struct db_attr_full *attrs = NULL;
	int i_id, i_name, i_created, i_last_update, i_rx_cnt, i_tx_cnt;
	int i_unit, i_txt, i_ts;
	char addf[512];
	char query[1024];
	char *qp[1] = {tok->t_name};
	PGresult *r;

	if (table && !valid_hist_table(table))
		return -1;

	if (tok && !(tok->t_flags & DB_T_WILDCARD)) {
		if (tok->t_name)
			snprintf(addf, sizeof(addf), " AND a.name = $1");
		else
			snprintf(addf, sizeof(addf), " AND a.id = %d", tok->t_id);
	} else
		memset(addf, 0, sizeof(addf));

	if (params && params->hp_flags & HP_HAS_FROM) {
		char t[64], d[64];
		strftime(d, sizeof(d), "%Y-%m-%d %H:%M:%S",
			 &params->hp_from.d_tm);
		snprintf(t, sizeof(t), " AND h.ts >= TIMESTAMP '%s.%u'",
			 d, params->hp_from.d_usec);
		xstrncat(addf, t, sizeof(addf));
	}

	if (params && params->hp_flags & HP_HAS_TO) {
		char t[64], d[64];
		strftime(d, sizeof(d), "%Y-%m-%d %H:%M:%S",
			 &params->hp_to.d_tm);
		snprintf(t, sizeof(t), " AND h.ts <= TIMESTAMP '%s.%u'",
			 d, params->hp_to.d_usec);
		xstrncat(addf, t, sizeof(addf));
	}

	snprintf(query, sizeof(query),
		 "SELECT a.id, a.name, h.rx_cnt, h.tx_cnt, " \
		 "h.ts, a.created, a.last_update, d.unit, d.txt " \
		 "FROM attrs AS a, attr_desc AS d, hist_%c AS h " \
		 "WHERE d.id = a.name AND h.attr = a.id AND " \
		 "a.item = %d %s ORDER BY a.item, a.name, h.ts",
		 table ? tolower(table[0]) : 'h', item_id, addf);

	r = xPQexecParams(conn, (tok && !(tok->t_flags & DB_T_WILDCARD) &&
				 tok->t_name) ? 1 : 0, qp, query);

	if (PQresultStatus(r) != PGRES_TUPLES_OK) {
		xwarn("Could not fetch attributes: %s", PQerrorMessage(conn));
		goto errout;
	}
	
	attrs = xcalloc(PQntuples(r), sizeof(struct db_attr_full));
	if (attrs == NULL)
		goto errout;

	col->size = PQntuples(r);
	col->array = attrs;
	
	i_id = PQfnumber(r, "id");
	i_name = PQfnumber(r, "name");
	i_rx_cnt = PQfnumber(r, "rx_cnt");
	i_tx_cnt = PQfnumber(r, "tx_cnt");
	i_unit = PQfnumber(r, "unit");
	i_txt = PQfnumber(r, "txt");
	i_created = PQfnumber(r, "created");
	i_last_update = PQfnumber(r, "last_update");
	i_ts = PQfnumber(r, "ts");
	
	for (i = 0; i < PQntuples(r); i++) {
		attrs[i].index = i;
		attrs[i].item = item_id;
		attrs[i].db_id = db_get_int(r, i, i_id);
		attrs[i].unit = db_get_int(r, i, i_unit);
		attrs[i].rx_cnt = db_get_int(r, i, i_rx_cnt);
		attrs[i].tx_cnt = db_get_int(r, i, i_tx_cnt);
		strncpy(attrs[i].name, db_get_charp(r, i, i_name), DB_ATTRLEN-1);
		strncpy(attrs[i].created, db_get_timestamp(r, i, i_created), DB_TSLEN-1);
		strncpy(attrs[i].last_update, db_get_timestamp(r, i, i_last_update), DB_TSLEN-1);
		strncpy(attrs[i].hts, db_get_timestamp(r, i, i_ts), DB_TSLEN-1);
		strncpy(attrs[i].txt, db_get_charp(r, i, i_txt), DB_TXTLEN-1);
	}
	
	err = 0;

errout:
	PQclear(r);

	return err;
}

int bdb_begin(PGconn *conn)
{
	int err = 0;
	PGresult *r = PQexec(conn, "BEGIN");

	if (PQresultStatus(r) != PGRES_COMMAND_OK) {
		xwarn("Could not begin transaction: %s", PQerrorMessage(conn));
		err = -1;
	}

	PQclear(r);

	return err;
}

int bdb_commit(PGconn *conn)
{
	int err = 0;
	PGresult *r = PQexec(conn, "COMMIT");

	if (PQresultStatus(r) != PGRES_COMMAND_OK) {
		xwarn("Could not commit transaction: %s", PQerrorMessage(conn));
		err = -1;
	}

	PQclear(r);

	return err;
}

int bdb_analyze(PGconn *conn)
{
	int err = 0;
	PGresult *r = PQexec(conn, "ANALYZE");

	if (PQresultStatus(r) != PGRES_COMMAND_OK) {
		xwarn("Could not analyze: %s", PQerrorMessage(conn));
		err = -1;
	}

	PQclear(r);

	return err;
}

static inline void print_token(FILE *fd, struct db_token *tok)
{
	if (tok->t_name)
		fprintf(fd, "%s", tok->t_name);
	else
		fprintf(fd, "%d", tok->t_id);

	if (tok->t_flags & DB_T_ATTR)
		fprintf(fd, "<attr>");
}

void bdb_print_filter(FILE *fd, struct db_filter *filter)
{
	if (filter->f_node)
		print_token(fd, filter->f_node);

	if (filter->f_item) {
		fprintf(fd, ".");
		print_token(fd, filter->f_item);
	}

	if (filter->f_attr) {
		fprintf(fd, ".");
		print_token(fd, filter->f_attr);
	}

	if (filter->f_field)
		fprintf(fd, "@%s", filter->f_field);
}

void *db_filter__scan_string(const char *);
void db_filter__switch_to_buffer(void *);
int db_filter_parse(void);

struct db_filter * bdb_parse_filter(const char *buf)
{
	struct db_filter *f;
	struct db_token *tok, *t;

	void *state = db_filter__scan_string(buf);
	db_filter__switch_to_buffer(state);

	if (db_filter_parse()) {
		xwarn("Failed to parse filter \"%s\"\n", buf);
		return NULL;
	}

	tok = db_filter_out; /* generated by yacc */
	if (tok == NULL)
		return NULL;
	
	f = xcalloc(1, sizeof(*f));

	f->f_node = tok;

	if (!tok->t_next)
		goto out;
	tok = tok->t_next;

	if (tok->t_flags & DB_T_ATTR) {
		xwarn("Node may not contain an attribute field\n");
		goto errout;
	}

	f->f_item = tok;

	if (!tok->t_next)
		goto out;
	tok = tok->t_next;

	if (tok->t_flags & DB_T_ATTR) {
		if (!tok->t_name)
			BUG();
		f->f_field = tok->t_name;
		goto out;
	} else
		f->f_attr = tok;

	if (!tok->t_next)
		goto out;
	tok = tok->t_next;

	if (tok->t_flags & DB_T_ATTR) {
		if (!tok->t_name)
			BUG();
		f->f_field = tok->t_name;
	} else {
		xwarn("Unexpected additional token after attribute\n");
		goto errout;
	}

out:
	/* free unused tokens */
	for (tok = tok->t_next ; tok; tok = t) {
		t = tok->t_next;
		if (tok->t_name)
			free(tok->t_name);
		free(tok);
	}

	return f;

errout:
	free(f);
	f = NULL;
	tok = db_filter_out;
	goto out;
}

int bdb_parse_date(const char *str, xdate_t *dst)
{
	time_t now = time(NULL);
	struct tm *now_tm = localtime(&now);
	const char *next;
	char *p;
	struct tm backup;

	memset(dst, 0, sizeof(*dst));

	if (strchr(str, '-')) {
		next = strptime(str, "%Y-%m-%d", &dst->d_tm);
		if (next == NULL ||
		    (*next != '\0' && *next != ' '))
			goto invalid_date;
	} else {
		dst->d_tm.tm_mday  = now_tm->tm_mday;
		dst->d_tm.tm_mon   = now_tm->tm_mon;
		dst->d_tm.tm_year  = now_tm->tm_year;
		dst->d_tm.tm_wday  = now_tm->tm_wday;
		dst->d_tm.tm_yday  = now_tm->tm_yday;
		dst->d_tm.tm_isdst = now_tm->tm_isdst;
		next = str;
	}

	if (*next == '\0')
		return 0;

	while (*next == ' ')
		next++;

	if (*next == '.')
		goto read_us;

	/* Make a backup, we can't rely on strptime to not screw
	 * up what we've got so far. */
	memset(&backup, 0, sizeof(backup));
	memcpy(&backup, &dst->d_tm, sizeof(backup));

	next = strptime(next, "%H:%M:%S", &dst->d_tm);
	if (next == NULL ||
	    (*next != '\0' && *next != '.'))
		goto invalid_date;

	dst->d_tm.tm_mday  = backup.tm_mday;
	dst->d_tm.tm_mon   = backup.tm_mon;
	dst->d_tm.tm_year  = backup.tm_year;
	dst->d_tm.tm_wday  = backup.tm_wday;
	dst->d_tm.tm_yday  = backup.tm_yday;
	dst->d_tm.tm_isdst = backup.tm_isdst;

	if (*next == '\0')
		return 0;
read_us:
	if (*next == '.')
		next++;
	else
		BUG();

	dst->d_usec = strtoul(next, &p, 10);

	if (*p != '\0')
		goto invalid_date;
	
	return 0;

invalid_date:
	fprintf(stderr, "Invalid date \"%s\"\n", str);
	return -1;
}
