/*
 * in_db.c		   Database Input
 *
 * Copyright (c) 2001-2005 Thomas Graf <tgraf@suug.ch>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <bmon/bmon.h>
#include <bmon/input.h>
#include <bmon/node.h>
#include <bmon/item.h>
#include <bmon/db.h>
#include <bmon/utils.h>
#include <bmon/attr.h>

static char *c_host;
static char *c_username;
static char *c_dbname;
static char *c_password;
static int c_localonly;
static int c_restore;

static PGconn *db_conn;

static void db_read(void)
{
	struct node *local_node = get_local_node();
	struct db_cnt *cnts;
	int i, n, ncnts;
	int last_item = -1;
	struct item *it = NULL, *parent = NULL;

	cnts = bdb_get_counters(db_conn, &ncnts);
	if (cnts == NULL)
		return;

	for (i = 0; i < ncnts; i++) {
		struct db_cnt *cnt = &cnts[i];

		if (last_item != cnt->item_index) {
			if (it) {
				notify_update(it, NULL);
				increase_lifetime(it, 1);
			}

			if (cnt->item_parent >= 0) {
				/* oh hell, O(n**2) but doesn't hurt too much
				 * at the moment I guess. */
				for (n = 0; n < ncnts; n++) {
					struct db_cnt *p = &cnts[n];
					if (p->item_index == cnt->item_parent)
						parent = cnt->arg;
				}

				if (parent == NULL) {
					xwarn("Could not find parent for " \
					      "%s:%d\n", cnt->item_name,
					      cnt->item_handle);
					it = NULL;
					continue;
				}
			} else
				parent = NULL;
				

			it = lookup_item(local_node, cnt->item_name,
					 cnt->item_handle, parent);
			if (it == NULL)
				continue;

			cnt->arg = it;

			it->i_major_attr = cnt->item_major_attr;
			it->i_minor_attr = cnt->item_minor_attr;
			it->i_level = cnt->item_indent;

			last_item = cnt->item_index;
		}

		if (it == NULL)
			BUG();

		update_attr(it, attr_name2type(cnt->attr_type),
			    cnt->attr_rx_cnt, cnt->attr_tx_cnt,
			    cnt->attr_rx_cnt ? RX_PROVIDED : 0 |
			    cnt->attr_tx_cnt ? TX_PROVIDED : 0);
	}

	if (it) {
		notify_update(it, NULL);
		increase_lifetime(it, 1);
	}

	xfree(cnts);
}

static void print_help(void)
{
	printf(
	"db - database input" \
	"\n" \
	"  Reads statistics from procfs (/proc/net/dev)\n" \
	"  Author: Thomas Graf <tgraf@suug.ch>\n" \
	"\n" \
	"  Options:\n" \
	"    file=PATH	    Path to statistics file (default: /proc/net/dev)\n");
}

static void db_set_opts(tv_t *attrs)
{
	while (attrs) {
		if (!strcasecmp(attrs->type, "localonly"))
			c_localonly = 1;
		else if (!strcasecmp(attrs->type, "restore"))
			c_restore = 1;
		else if (!strcasecmp(attrs->type, "host") && attrs->value)
			c_host = attrs->value;
		else if (!strcasecmp(attrs->type, "username") && attrs->value)
			c_username = attrs->value;
		else if (!strcasecmp(attrs->type, "dbname") && attrs->value)
			c_dbname = attrs->value;
		else if (!strcasecmp(attrs->type, "password") && attrs->value)
			c_password = attrs->value;
		else if (!strcasecmp(attrs->type, "help")) {
			print_help();
			exit(0);
		}
		attrs = attrs->next;
	}
}

static int db_probe(void)
{
	if (!c_host)
		c_host = get_db_host();

	if (!c_username)
		c_username = get_db_user();

	if (!c_dbname)
		c_dbname = get_db_name();

	if (!c_password)
		c_password = get_db_password();

	if (c_password == NULL) {
		fprintf(stderr, "You must specify the database password\n");
		return 0;
	}

	db_conn = bdb_connect(c_host, c_username, c_password, c_dbname);
	if (db_conn == NULL)
		return 0;

	return 1;
}

static void db_shutdown(void)
{
	if (db_conn)
		bdb_shutdown(db_conn);
}

static struct input_module db_ops = {
	.im_name = "db",
	.im_read = db_read,
	.im_set_opts = db_set_opts,
	.im_probe = db_probe,
	.im_shutdown = db_shutdown,
};

static void __init db_init(void)
{
	register_secondary_input_module(&db_ops);
}
