/*
 * item.c		Item Management
 *
 * Copyright (c) 2001-2005 Thomas Graf <tgraf@suug.ch>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <bmon/bmon.h>
#include <bmon/node.h>
#include <bmon/conf.h>
#include <bmon/item.h>
#include <bmon/attr.h>
#include <bmon/itemtab.h>
#include <bmon/input.h>
#include <bmon/graph.h>
#include <bmon/utils.h>

#define MAX_POLICY             255
#define SECOND                 1.0f
#define MINUTE                60.0f
#define HOUR                3600.0f
#define DAY                86400.0f

static char * allowed_items[MAX_POLICY];
static char * denied_items[MAX_POLICY];
static char * allowed_attrs[MAX_POLICY];
static char * denied_attrs[MAX_POLICY];
static int allow_all_attrs;

static inline uint8_t attr_hash(int type)
{
	return (type & 0xFF) % ATTR_HASH_MAX;
}

struct stat_attr *lookup_attr(struct item *it, int type)
{
	struct stat_attr *a;
	uint8_t h = attr_hash(type);
	
	for (a = it->i_attrs[h]; a; a = a->a_next)
		if (a->a_type == type)
			return a;
	return NULL;
}

inline b_cnt_t attr_get_rx(struct stat_attr *a)
{
	return a->a_rx.r_total_sum;
}

inline rate_cnt_t attr_get_rx_rate(struct stat_attr *a)
{
	return a->a_rx.r_tps;
}

inline b_cnt_t attr_get_rx_last(struct stat_attr *a)
{
	return a->a_rx.r_current;
}

inline b_cnt_t attr_get_tx(struct stat_attr *a)
{
	return a->a_tx.r_total_sum;
}

inline rate_cnt_t attr_get_tx_rate(struct stat_attr *a)
{
	return a->a_tx.r_tps;
}

inline b_cnt_t attr_get_tx_last(struct stat_attr *a)
{
	return a->a_tx.r_current;
}

static int match_mask(const char *mask, const char *str)
{
	int i, n;
	char c;

	if (!mask || !str)
		return 0;
	
	for (i = 0, n = 0; mask[i] != '\0'; i++, n++) {
		if (mask[i] == '*') {
			c = tolower(mask[i+1]);
			
			if (c == '\0')
				return 1;
			
			for (; tolower(str[n]) != c; n++)
				if (str[n] == '\0')
					return 0;
			
			n--;
		} else if (tolower(mask[i]) != tolower(str[n]))
			return 0;
	}

	return str[n] == '\0' ? 1 : 0;
}

static int item_allowed(const char *name)
{
	int n;
	
	if (!allowed_items[0] && !denied_items[0])
		return 1;

	if (!allowed_items[0]) {
		for (n = 0; n < MAX_POLICY && denied_items[n]; n++)
			if (match_mask(denied_items[n], name))
				return 0;
	
		return 1;
	}

	for (n = 0; n < MAX_POLICY && denied_items[n]; n++)
		if (match_mask(denied_items[n], name))
			return 0;
	
	for (n=0; n < MAX_POLICY && allowed_items[n]; n++)
		if (match_mask(allowed_items[n], name))
			return 1;

	return 0;
}

void item_parse_policy(const char *policy)
{
	static int set = 0;
	int i, a = 0, d = 0, f = 0;
	char *p, *s;

	if (set)
		return;
	set = 1;
	
	s = strdup(policy);
	
	for (i = 0, p = s; ; i++) {
		if (s[i] == ',' || s[i] == '\0') {

			f = s[i] == '\0' ? 1 : 0;
			s[i] = '\0';
			
			if ('!' == *p) {
				if (d > (MAX_POLICY - 1))
					break;
				denied_items[d++] = strdup(++p);
			} else {
				if(a > (MAX_POLICY - 1))
					break;
				allowed_items[a++] = strdup(p);
			}
			
			if (f)
				break;
			
			p = &s[i+1];
		}
	}
	
	xfree(s);
}

static inline int collect_history(struct item *item, int type)
{
	int n;
	const char *name = attr_type2name(type);

	if (item->i_major_attr == type ||
	    item->i_minor_attr == type)
		return 1;

	if (allow_all_attrs)
		return 1;

	if (!allowed_attrs[0] && !denied_attrs[0]) {
		if (!strcmp(name, "bytes") || !strcmp(name, "packets"))
			return 1;
		else
			return 0;
	}

	if (!allowed_attrs[0]) {
		for (n = 0; n < MAX_POLICY && denied_attrs[n]; n++)
			if (!strcasecmp(denied_attrs[n], name))
				return 0;
		return 1;
	}

	for (n = 0; n < MAX_POLICY && denied_attrs[n]; n++)
		if (!strcasecmp(denied_attrs[n], name))
			return 0;
	
	for (n=0; n < MAX_POLICY && allowed_attrs[n]; n++)
		if (!strcasecmp(allowed_attrs[n], name))
			return 1;

	return 0;
}

void parse_attr_policy(const char *policy)
{
	static int set = 0;
	int i, a = 0, d = 0, f = 0;
	char *p, *s;

	if (set)
		return;
	set = 1;

	if (!strcasecmp(policy, "all")) {
		allow_all_attrs = 1;
		return ;
	}

	s = strdup(policy);
	
	for (i = 0, p = s; ; i++) {
		if (s[i] == ',' || s[i] == '\0') {

			f = s[i] == '\0' ? 1 : 0;
			s[i] = '\0';
			
			if ('!' == *p) {
				if (d > (MAX_POLICY - 1))
					break;
				denied_attrs[d++] = strdup(++p);
			} else {
				if(a > (MAX_POLICY - 1))
					break;
				allowed_attrs[a++] = strdup(p);
			}
			
			if (f)
				break;
			
			p = &s[i+1];
		}
	}
	
	xfree(s);
}

struct item * search_item(struct item_group *group, const char *name, uint32_t handle,
			  struct item *parent)
{
	struct item *item;
	
	if (group == NULL)
		BUG();

	for (item = group->g_items; item; item = item->i_next)
		if (!strcmp(name, item->i_name) &&
		    item->i_handle == handle &&
		    (!parent || item->i_parent == parent))
			return item;

	return NULL;
}

struct item *__lookup_item(struct item_group *group, const char *name,
			   uint32_t handle, struct item *parent)
{
	struct item *item;

	if (parent) {
		for (item = parent->i_childs; item; item = item->i_next)
			if (!strcmp(name, item->i_name) && item->i_handle == handle)
				return item;
	} else {
		for (item = group->g_items; item; item = item->i_next)
			if (!strcmp(name, item->i_name) && item->i_handle == handle)
				return item;
	}

	return NULL;
}


struct item * lookup_item(struct node *node, struct item_group *group,
			  const char *name, uint32_t handle, struct item *parent)
{
	struct item *item;
	struct it_item *tab;
	
	if (!node || !group)
		BUG();

	item = __lookup_item(group, name, handle, parent);

	if (item)
		return !(item->i_flags & ITEM_FLAG_UPDATED) ? item : NULL;

	if (!handle && !item_allowed(name))
		return NULL;

	item = xcalloc(1, sizeof(*item));

	strncpy(item->i_name, name, sizeof(item->i_name) - 1);
	item->i_handle = handle;
	item->i_parent = parent;
	item->i_node = node;
	item->i_group = group;
	item->i_lifetime = get_lifetime();

	if (!strcmp(node->n_name, get_local_nodename()))
		item->i_flags |= ITEM_FLAG_LOCAL;

	tab = lookup_tab(item);

	if (tab && tab->ii_desc)
		item->i_desc = tab->ii_desc;

	if (parent) {
		item->i_prev = parent->i_childs_tail;

		if (parent->i_childs_tail)
			parent->i_childs_tail->i_next = item;
		else
			parent->i_childs = item;

		parent->i_childs_tail = item;
	} else {
		item->i_prev = group->g_items_tail;

		if (group->g_items_tail)
			group->g_items_tail->i_next = item;
		else
			group->g_items = item;

		group->g_items_tail = item;
	}

	group->g_nitems++;

	return item;
}

void foreach_child(struct item_group *group, struct item *parent,
		   void (*cb)(struct item *, void *), void *arg)
{
	struct item *item;

	for (item = parent->i_childs; item; item = item->i_next)
		cb(item, arg);
}

void reset_item(struct item *i)
{
	i->i_flags &= ~ITEM_FLAG_UPDATED;
}

static inline void unlink_item(struct item *item)
{
	if (item->i_prev)
		item->i_prev->i_next = item->i_next;

	if (item->i_next)
		item->i_next->i_prev = item->i_prev;

	if (item->i_parent) {
		if (!item->i_prev)
			item->i_parent->i_childs = item->i_next;

		if (!item->i_next)
			item->i_parent->i_childs_tail = item->i_prev;

	} else {
		if (!item->i_prev)
			item->i_group->g_items = item->i_next;

		if (!item->i_next)
			item->i_group->g_items_tail = item->i_prev;
	}
}

static inline void delete_item(struct item *item)
{
	int m;

	for (m = 0; m < ATTR_HASH_MAX; m++) {
		struct stat_attr *a, *next;
		for (a = item->i_attrs[m]; a; a = next) {
			next = a->a_next;
			xfree(a);
		}
	}

	if (item->i_group->g_selected == item) {
		if (prev_item() == END_OF_LIST) {
			if (prev_group() == END_OF_LIST) {
				if (prev_node() != END_OF_LIST)
					last_item();
			} else
				last_item();
		}
	}

	unlink_item(item);
	item->i_group->g_nitems--;

	if (item->i_childs) {
		struct item *child;

		for (child = item->i_childs; child; child = child->i_next)
			delete_item(child);
	}
	
	xfree(item);
}


void remove_unused_items(struct item *item)
{
	if (--(item->i_lifetime) <= 0)
		delete_item(item);
}

static void calc_rate(struct rate *rate, timestamp_t *ts, int ignore_overflows)
{
	float diff;
	b_cnt_t delta;

	if (get_known_reboot()) {
		/* We know that reboot must have happened and thus the
		 * counter itself is the delta. This makes sure we
		 * take the period of time between the interface comming
		 * up and the bmon instance starting up into account. */
		delta = rate->r_current;
	} else if (rate->r_current < rate->r_prev) {
		/* We are likely to have hit an overflow, we assume overflows
		 * only happen on 32bit counters for now.
		 *
		 * NOTE: In case we restored from a statefile we can only
		 * detect one overflow, additional overflows will be
		 * silently ignored. */
		if (ignore_overflows)
			delta = rate->r_current;
		else
			delta = (OVERFLOW_LIMIT-rate->r_prev)+rate->r_current;
	} else
		delta = rate->r_current - rate->r_prev;

	rate->r_total_sum += delta;
	rate->r_prev = rate->r_current;

	if (!rate->r_total_prev) {
		rate->r_total_prev = rate->r_total_sum;
		COPY_TS(&rate->r_last_update, ts);
		return;
	}
	
	diff = time_diff(&rate->r_last_update, ts);
	
	if (diff >= get_rate_interval()) {
		rate_cnt_t old_rate = rate->r_tps;
		b_cnt_t rate_delta = rate->r_total_sum - rate->r_total_prev;

		rate->r_tps = rate_delta / diff;
		rate->r_tps = ((rate->r_tps * 3) + old_rate) / 4;
		COPY_TS(&rate->r_last_update, ts);
		rate->r_total_prev = rate->r_total_sum;
	}
}

static inline void update_history_data(struct hist_data *hd, struct rate *r,
				       int index, double diff, int is_rate)
{
	if (is_rate)
		hd->hd_data[index] = r->r_total_sum / diff;
	else {
		hd->hd_data[index] = (r->r_total_sum - hd->hd_prev_total)/diff;
		hd->hd_prev_total = r->r_total_sum;
	}
}

static void update_history_element(struct hist_elem *he, struct rate *rx,
				   struct rate *tx, timestamp_t *ts, float unit,
				   int is_rate)
{
	double diff = time_diff(&he->he_last_update, ts);
	
	if (!he->he_last_update.tv_sec)
		diff = 0.0f;

	if (get_read_interval() > unit)
		goto discard;

	/* Discard outdated updates */
	if (diff >= (unit + (get_hb_factor() * unit)))
		goto discard;

	if (!he->he_last_update.tv_sec && !is_rate) {
		he->he_rx.hd_prev_total = rx->r_total_sum;
		he->he_tx.hd_prev_total = tx->r_total_sum;
	} else if (get_read_interval() == unit) {
		update_history_data(&he->he_rx, rx, he->he_index,
				    1.0f, is_rate);
		update_history_data(&he->he_tx, tx, he->he_index,
				    1.0f, is_rate);
	/* The timing code might do shorter intervals than requested to
	 * adjust previous intervals being too long */
	} else if (diff >= ((1.0f - get_hb_factor()) * unit)) {
		update_history_data(&he->he_rx, rx, he->he_index,
				    diff, is_rate);
		update_history_data(&he->he_tx, tx, he->he_index,
				    diff, is_rate);
	} else {
		/* Silently discard and give it a chance to come back
		 * in time. */
		return;
	}

	if (he->he_index >= (HISTORY_SIZE - 1))
		he->he_index = 0;
	else
		he->he_index++;

	COPY_TS(&he->he_last_update, ts);

	return;
discard:
	while(diff > 0.0f) {
		he->he_rx.hd_data[he->he_index] = UNK_DATA;
		he->he_tx.hd_data[he->he_index] = UNK_DATA;

		if (he->he_index >= (HISTORY_SIZE - 1))
			he->he_index = 0;
		else
			he->he_index++;

		diff -= unit;
	}

	he->he_rx.hd_prev_total = rx->r_total_sum;
	he->he_tx.hd_prev_total = tx->r_total_sum;
	COPY_TS(&he->he_last_update, ts);
}

static inline void update_history(struct stat_attr_hist *ah, timestamp_t *ts)
{
	struct history *hist = &ah->a_hist;
	struct rate *rx = &ah->a_rx;
	struct rate *tx = &ah->a_tx;
	int is_rate = ah->a_flags & ATTR_FLAG_IS_RATE;
	
	if (get_read_interval() != 1.0f)
		update_history_element(&hist->h_read, rx, tx, ts,
				       get_read_interval(), is_rate);

	update_history_element(&hist->h_sec, rx, tx, ts, SECOND, is_rate);
	update_history_element(&hist->h_min, rx, tx, ts, MINUTE, is_rate);
	update_history_element(&hist->h_hour, rx, tx, ts, HOUR, is_rate);
	update_history_element(&hist->h_day, rx, tx, ts, DAY, is_rate);
}

static void update_attr_hist(struct stat_attr *a, void *arg)
{
	timestamp_t *ts = (timestamp_t *) arg;
	int ign_of = a->a_flags & ATTR_FLAG_IGNORE_OVERFLOWS;

	if (ts == NULL)
		ts = &rtiming.rt_last_read;

	if (a->a_flags & ATTR_FLAG_IS_RATE) {
		a->a_rx.r_tps = a->a_rx.r_total_sum = a->a_rx.r_current;
		a->a_tx.r_tps = a->a_tx.r_total_sum = a->a_tx.r_current;
	} else {
		calc_rate(&a->a_rx, ts, ign_of);
		calc_rate(&a->a_tx, ts, ign_of);
	}

	if (a->a_flags & ATTR_FLAG_HISTORY)
		update_history((struct stat_attr_hist *) a, ts);
}

static inline float calc_usage(rate_cnt_t rate, b_cnt_t max)
{
	if (!max)
		return -1.0f;

	if (!rate)
		return 0.0f;

	return 100.0f / ((double) max / (double) rate);
}

void notify_update(struct item *i, timestamp_t *ts)
{
	struct it_item *tab;
	struct stat_attr *a;

	i->i_flags |= ITEM_FLAG_UPDATED;
	foreach_attr(i, &update_attr_hist, ts);
	
	a = lookup_attr(i, i->i_usage_attr);
	tab = lookup_tab(i);

	if (a && a->a_flags & ATTR_FLAG_IS_USAGE) {
		i->i_rx_usage = attr_get_rx_rate(a);
		i->i_tx_usage = attr_get_tx_rate(a);
	} else if (a && tab) {
		i->i_rx_usage = calc_usage(attr_get_rx_rate(a), tab->ii_rx_max);
		i->i_tx_usage = calc_usage(attr_get_tx_rate(a), tab->ii_tx_max);
	} else if (!i->i_node || !i->i_node->n_from) {
		i->i_rx_usage = -1;
		i->i_tx_usage = -1;
	}
}

void increase_lifetime(struct item *i, int l)
{
	i->i_lifetime = l*get_lifetime();
}

void foreach_attr(struct item *i,
		  void (*cb)(struct stat_attr *, void *), void *arg)
{
	int m;
	struct stat_attr *a;

	for (m = 0; m < ATTR_HASH_MAX; m++)
		for (a = i->i_attrs[m]; a; a = a->a_next)
			cb(a, arg);
}


void update_attr(struct item *i, int type, b_cnt_t rx, b_cnt_t tx, int flags)
{
	struct stat_attr *a = lookup_attr(i, type);

	if (a == NULL) {
		uint8_t h = attr_hash(type);

		if (collect_history(i, type)) {
			a = xcalloc(1, sizeof(struct stat_attr_hist));
			a->a_flags |= ATTR_FLAG_HISTORY;
		} else
			a = xcalloc(1, sizeof(struct stat_attr));

		a->a_type = type;
		a->a_unit = attr_unit(type);
		if (a->a_unit == NULL)
			BUG();

		if (attr_ignore_overflows(type))
			a->a_flags |= ATTR_FLAG_IGNORE_OVERFLOWS;

		if (attr_is_rate(type))
			a->a_flags |= ATTR_FLAG_IS_RATE;

		if (attr_is_usage(type))
			a->a_flags |= ATTR_FLAG_IS_USAGE;

		if (flags & IS64BIT)
			a->a_flags |= ATTR_FLAG_IS_64BIT;

		a->a_next = i->i_attrs[h];
		i->i_attrs[h] = a;
		i->i_nattrs++;
	}

	if (flags & RX_PROVIDED) {
		a->a_rx.r_current = rx;
		a->a_flags |= ATTR_FLAG_RX_ENABLED;
		update_ts(&a->a_updated);
	}

	if (flags & TX_PROVIDED) {
		a->a_tx.r_current = tx;
		a->a_flags |= ATTR_FLAG_TX_ENABLED;
		update_ts(&a->a_updated);
	}
}

void restore_attr(struct item *i, int type, b_cnt_t rx_sum, b_cnt_t tx_sum)
{
	struct stat_attr *a = lookup_attr(i, type);

	if (a == NULL)
		BUG();

	/* Restore the old total sum, the missed delta will be
	 * added after the first read.
	 *
	 * NOTE: We do not restore the r_total_prev counter to
	 * avoid calculating a completely bogus rate, the
	 * starting period of rate calculations is postponed
	 * to after the first read. */
	a->a_rx.r_total_sum = rx_sum;
	a->a_tx.r_total_sum = tx_sum;

	/* Assign the current rate (:= last counter state of old
	 * bmon instance) as prev counter state so we can calculate
	 * the delta of the missed values. */
	a->a_rx.r_prev = a->a_rx.r_current;
	a->a_tx.r_prev = a->a_tx.r_current;
}

int __first_attr(struct item *item, int graph)
{
	int i;
	struct stat_attr *a;

	for (i = 0; i < ATTR_HASH_MAX; i++) {
		for (a = item->i_attrs[i]; a; a = a->a_next) {
			if (a->a_flags & ATTR_FLAG_HISTORY) {
				item->i_attr_sel[graph] = a;
				return 0;
			}
		}
	}

	return EMPTY_LIST;
}

int __next_attr(struct item *item, int graph)
{
	int hash;
	struct stat_attr *attr, *next;

	if (item->i_attr_sel[graph] == NULL)
		return __first_attr(item, graph);

	attr = item->i_attr_sel[graph];
	hash = attr_hash(attr->a_type);
	next = attr->a_next;

	if (next == NULL)
		hash++;

	for (; hash < ATTR_HASH_MAX; hash++) {
		if (next) {
			attr = next;
			next = NULL;
		} else
			attr = item->i_attrs[hash];

		for (; attr; attr = attr->a_next) {
			if (!(attr->a_flags & ATTR_FLAG_HISTORY))
				continue;
			item->i_attr_sel[graph] = attr;
			return 0;
		}
	}

	return __first_attr(item, graph);
}

int first_attr(void)
{
	struct item *item = get_current_item();
	
	if (item == NULL)
		return EMPTY_LIST;

	return __first_attr(item, item->i_graph_sel);
}

int next_attr(void)
{
	struct item *item = get_current_item();
	
	if (item == NULL)
		return EMPTY_LIST;

	return __next_attr(item, item->i_graph_sel);
}

struct stat_attr *current_attr(struct item *item, int graph)
{
	if (item->i_attr_sel[graph] == NULL)
		__first_attr(item, graph);

	return item->i_attr_sel[graph];
}

int set_graph_unit(int unit)
{
	struct item *item = get_current_item();
	if (item == NULL)
		return EMPTY_LIST;

	item->i_unit[item->i_graph_sel] = unit;

	return 0;
}
