/*
 * node.c            node handling
 *
 * Copyright (c) 2001-2005 Thomas Graf <tgraf@suug.ch>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <bmon/bmon.h>
#include <bmon/node.h>
#include <bmon/item.h>
#include <bmon/group.h>
#include <bmon/utils.h>

static struct node *nodes;
static size_t nodes_size;
static int nnodes;
static int ngroups;

static const char * local_node_name;
static struct node *local_node;
static struct node *current_node;


struct node * lookup_node(const char *name, int creat)
{
	int i;

	if (NULL == nodes) {
		nodes_size = 32;
		nodes = xcalloc(nodes_size, sizeof(struct node));
	}

	for (i = 0; i < nodes_size; i++)
		if (nodes[i].n_name && !strcmp(name, nodes[i].n_name))
			return &nodes[i];

	if (creat) {
		for (i = 0; i < nodes_size; i++)
			if (NULL == nodes[i].n_name)
				break;

		while (i >= nodes_size) {
			int oldsize = nodes_size;
			nodes_size += 32;
			nodes = xrealloc(nodes, nodes_size * sizeof(struct node));
			memset((uint8_t *) nodes + oldsize, 0, (nodes_size - oldsize) * sizeof(struct node));
		}

		nodes[i].n_name = strdup(name);
		nodes[i].n_index = i;
		nnodes++;
		return &nodes[i];
	}

	return NULL;
}

void foreach_node(void (*cb)(struct node *, void *), void *arg)
{
	int i;

	for (i = 0; i < nnodes; i++)
		cb(&nodes[i], arg);
}

struct node *get_node(int index)
{
	return &nodes[index];
}

void foreach_group(struct node *n, void (*cb)(struct item_group *, void *), void *arg)
{
	struct item_group *g;

	for (g = n->n_groups; g; g = g->g_next)
		cb(g, arg);
}

static inline void __foreach_item(struct item *root,
				  void (*cb)(struct item *, void *), void *arg)
{
	struct item *item;

	for (item = root; item; item = item->i_next) {
		cb(item, arg);
		if (item->i_childs)
			__foreach_item(item->i_childs, cb, arg);
	}
}


void foreach_item(struct item_group *g, void (*cb)(struct item *, void *), void *arg)
{
	__foreach_item(g->g_items, cb, arg);
}

static inline void __foreach_item_deep(struct item *root, struct node *node,
				       struct item_group *group,
				       void (*cb)(struct node *, struct item_group *,
						  struct item *, void *),
				       void *arg)
{
	struct item *item;

	for (item = root; item; item = item->i_next) {
		cb(node, group, item, arg);
		if (item->i_childs)
			__foreach_item_deep(item->i_childs, node, group, cb, arg);
	}
}

void foreach_item_deep(void (*cb)(struct node *, struct item_group *, struct item *,
				  void *), void *arg)
{
	int i;

	for (i = 0; i < nnodes; i++) {
		struct item_group *g;
		struct node *n = &nodes[i];
		
		for (g = n->n_groups; g; g = g->g_next)
			__foreach_item_deep(g->g_items, n, g, cb, arg);
	}
}


static void __reset_item(struct node *n, struct item_group *g, struct item *i, void *arg)
{
	reset_item(i);
}

void reset_nodes(void)
{
	foreach_item_deep(__reset_item, NULL);
}

static void __remove_unused_item(struct node *n, struct item_group *g,
				 struct item *i, void *arg)
{
	remove_unused_items(i);
}

void remove_unused_node_items(void)
{
	foreach_item_deep(__remove_unused_item, NULL);
}

struct node * get_local_node(void)
{
	if (NULL == local_node)
		local_node = lookup_node(local_node_name, 1);

	return local_node;
}

const char * get_local_nodename(void)
{
	return local_node_name;
}

int get_nnodes(void)
{
	return nnodes;
}

int get_ngroups(void)
{
	return ngroups;
}

static void get_node_info(void)
{
	struct utsname uts;
	
	if (uname(&uts) < 0)
		quit("utsname failed: %s\n", strerror(errno));
	
	local_node_name = strdup(uts.nodename);
}

static void __calc_item_rate(struct item *item, void *arg)
{
	struct item_group *group = arg;

	if (item->i_level == 0 && !(item->i_flags & ITEM_FLAG_EXCLUDE)) {
		struct stat_attr *maj = lookup_attr(item, item->i_major_attr);
		struct stat_attr *min = lookup_attr(item, item->i_minor_attr);

		if (maj) {
			group->g_rx_maj_total += attr_get_rx_rate(maj);
			group->g_tx_maj_total += attr_get_tx_rate(maj);
			group->g_maj_unit = maj->a_unit;
		}

		if (min) {
			group->g_rx_min_total += attr_get_rx_rate(min);
			group->g_tx_min_total += attr_get_tx_rate(min);
			group->g_min_unit = min->a_unit;
		}
		
	}
}

static void __calc_group_rate(struct item_group *group, void *arg)
{
	group->g_rx_maj_total = 0;
	group->g_tx_maj_total = 0;
	group->g_rx_min_total = 0;
	group->g_tx_min_total = 0;

	foreach_item(group, __calc_item_rate, group);
}

void calc_node_rate(struct node *node)
{
	foreach_group(node, __calc_group_rate, NULL);
}

void calc_node_rates(void)
{
	int i;

	for (i = 0; i < nnodes; i++)
		calc_node_rate(&nodes[i]);
}

struct node * get_current_node(void)
{
	return current_node;
}

struct item_group *get_current_group(struct node *node)
{
	return node->n_current_group;
}

struct item * get_current_item(void)
{
	struct item_group *group;

	if (!current_node)
		return NULL;

	group = get_current_group(current_node);
	if (group == NULL)
		return NULL;

	return group->g_selected;
}

struct item_group *get_group(struct node *node, const char *name, int creat)
{
	struct item_group *g;
	struct group_data *gd;

	for (g = node->n_groups; g; g = g->g_next)
		if (!strcmp(name, g->g_name))
			return g;

	if (!creat)
		return NULL;

	gd = lookup_group_data(name);
	if (gd == NULL) {
		xwarn("Cannot find group definition for group \"%s\"\n", name);
		return NULL;
	}

	g = xcalloc(1, sizeof(*g));
	g->g_name = gd->gd_name;
	g->g_data = gd;

	g->g_prev = node->n_groups_tail;
	g->g_next = NULL;

	if (node->n_groups_tail)
		node->n_groups_tail->g_next = g;
	else
		node->n_groups = g;
	
	node->n_groups_tail = g;

	node->n_ngroups++;
	ngroups++;

	return g;
}

int first_node(void)
{
	if (nnodes <= 0)
		return EMPTY_LIST;

	current_node = &nodes[0];
	return 0;
}

int last_node(void)
{
	if (nnodes <= 0)
		return EMPTY_LIST;

	current_node = &nodes[nnodes - 1];
	return 0;
}

int next_node(void)
{
	if (nnodes <= 0)
		return EMPTY_LIST;
	
	if (current_node == NULL)
		return first_node();

	if (current_node->n_index >= (nnodes - 1))
		return END_OF_LIST;

	current_node = &nodes[current_node->n_index + 1];
	return 0;
}

int prev_node(void)
{
	if (nnodes <= 0)
		return EMPTY_LIST;
	
	if (current_node == NULL)
		return last_node();

	if (current_node->n_index <= 0)
		return END_OF_LIST;

	current_node = &nodes[current_node->n_index - 1];
	return 0;
}

int first_group(void)
{
	if (current_node == NULL) {
		first_node();
		if (current_node == NULL)
			return EMPTY_LIST;
	}
	
	if (current_node->n_ngroups <= 0)
		return EMPTY_LIST;

	current_node->n_current_group = current_node->n_groups;
	return 0;
}

int last_group(void)
{
	if (current_node == NULL) {
		first_node();
		if (current_node == NULL)
			return EMPTY_LIST;
	}
	
	if (current_node->n_ngroups <= 0)
		return EMPTY_LIST;

	current_node->n_current_group = current_node->n_groups_tail;
	return 0;
}

int next_group(void)
{
	if (current_node == NULL) {
		first_node();
		if (current_node == NULL)
			return EMPTY_LIST;
	}

	if (current_node->n_current_group == NULL)
		return first_group();

	if (!current_node->n_current_group->g_next)
		return END_OF_LIST;

	current_node->n_current_group = current_node->n_current_group->g_next;
	return 0;
}

int prev_group(void)
{
	if (current_node == NULL) {
		first_node();
		if (current_node == NULL)
			return EMPTY_LIST;
	}

	if (current_node->n_current_group == NULL)
		return last_group();

	if (!current_node->n_current_group->g_prev)
		return END_OF_LIST;

	current_node->n_current_group = current_node->n_current_group->g_prev;
	return 0;
}

int first_item(void)
{
	if (current_node == NULL) {
		first_node();
		if (current_node == NULL)
			return EMPTY_LIST;
	}

	if (current_node->n_current_group == NULL) {
		first_group();
		if (current_node->n_current_group == NULL)
			return EMPTY_LIST;
	}

	if (current_node->n_current_group->g_nitems <= 0)
		return EMPTY_LIST;

	if (!current_node->n_current_group->g_items)
		BUG();

	current_node->n_current_group->g_selected =
		current_node->n_current_group->g_items;
	return 0;
}

int last_item(void)
{
	if (current_node == NULL) {
		first_node();
		if (current_node == NULL)
			return EMPTY_LIST;
	}

	if (current_node->n_current_group == NULL) {
		first_group();
		if (current_node->n_current_group == NULL)
			return EMPTY_LIST;
	}

	if (current_node->n_current_group->g_nitems <= 0)
		return EMPTY_LIST;

	if (!current_node->n_current_group->g_items_tail)
		BUG();

	current_node->n_current_group->g_selected =
		current_node->n_current_group->g_items_tail;
	return 0;
}

int next_item(void)
{
	struct item *item;
	struct item_group *group;

	if (current_node == NULL) {
		first_node();
		if (current_node == NULL)
			return EMPTY_LIST;
	}

	if (current_node->n_current_group == NULL) {
		first_group();
		if (current_node->n_current_group == NULL)
			return EMPTY_LIST;
	}

	group = current_node->n_current_group;

	if (group->g_nitems <= 0)
		return EMPTY_LIST;

	if (!group->g_selected)
		return first_item();

	item = group->g_selected;

	if (item->i_childs && !item_folded(item)) {
		group->g_selected = item->i_childs;
		return 0;
	} else if (item->i_next) {
		group->g_selected = item->i_next;
		return 0;
	} 

	while (item->i_parent && !item->i_next)
		item = item->i_parent;

	if (item->i_next) {
		group->g_selected = item->i_next;
		return 0;
	} else
		return END_OF_LIST;
}

int prev_item(void)
{
	struct item *item;
	struct item_group *group;

	if (current_node == NULL) {
		first_node();
		if (current_node == NULL)
			return EMPTY_LIST;
	}

	if (current_node->n_current_group == NULL) {
		first_group();
		if (current_node->n_current_group == NULL)
			return EMPTY_LIST;
	}

	group = current_node->n_current_group;

	if (group->g_nitems <= 0)
		return EMPTY_LIST;

	if (!group->g_selected)
		return last_item();

	item = group->g_selected;

	if (item->i_prev) {
		item = item->i_prev;
		while (item->i_childs_tail && !item_folded(item))
			item = item->i_childs_tail;
		group->g_selected = item;
		return 0;
	} else if (item->i_parent) {
		group->g_selected = item->i_parent;
		return 0;
	} else
		return END_OF_LIST;
}

static void __init node_init(void)
{
	get_node_info();
}
