/*
 * out_db.c		Database Output
 *
 * Copyright (c) 2001-2005 Thomas Graf <tgraf@suug.ch>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <bmon/bmon.h>
#include <bmon/node.h>
#include <bmon/output.h>
#include <bmon/graph.h>
#include <bmon/input.h>
#include <bmon/utils.h>
#include <bmon/db.h>

static char *c_host;
static char *c_username;
static char *c_dbname;
static char *c_password;
static int   c_interval = 3;

static PGconn *db_conn;

static inline int cur_hist_index(int index)
{
	return index == 0 ? (HISTORY_SIZE - 1) : (index - 1);
}

static void write_per_attr(struct stat_attr *a, void *arg)
{
	int item_id = *((int *) arg);
	int attr_id = bdb_find_attr(a, item_id);

	if (attr_id < 0) {
		attr_id = bdb_insert_attr(db_conn, a, item_id);
		if (attr_id < 0)
			return;
	} else if (bdb_update_attr(db_conn, a, attr_id) < 0)
			return;
}

struct xdata {
	int node_id;
	struct node *node;
};

static void write_per_item(struct item *item, void *arg)
{
	struct xdata *x = arg;
	int id = bdb_find_item(item, x->node, x->node_id);

	if (id < 0) {
		id = bdb_insert_item(db_conn, item, x->node, x->node_id);
		if (id < 0)
			return;
	}

        if (item->i_rx_usage >= 0 ||
            item->i_tx_usage >= 0)
                bdb_update_item(db_conn, item, id);
	foreach_attr(item, &write_per_attr, &id);
}

static void write_per_node(struct node *node, void *arg)
{
	struct xdata x = {
		.node = node,
		.node_id = bdb_find_node(node),
	};

	if (x.node_id < 0) {
		x.node_id = bdb_insert_node(db_conn, node);
		if (x.node_id < 0)
			return;
	}

	foreach_item(node, write_per_item, &x);
}

void db_draw(void)
{
	static int rem = 1;

	if (--rem)
		return;
	else
		rem = c_interval;

	if (bdb_begin(db_conn) < 0)
		return;

	if (bdb_update_caches(db_conn) < 0)
		return;

	foreach_node(write_per_node, NULL);

	bdb_finish_updates(db_conn);

	bdb_commit(db_conn);
}

static void print_module_help(void)
{
	printf(
	"DB - Database Output\n" \
	"\n" \
	"  Writes current rate estimations into a database for\n" \
	"  other tools to pick up.\n" \
	"\n" \
	"  Author: Thomas Graf <tgraf@suug.ch>\n" \
	"\n" \
	"  Options:\n" \
	"    host=HOST        Host the database is on (default: localhost)\n" \
	"    dbname=NAME      Name of database (default: bmon)\n" \
	"    username=NAME    Authentication username (default: bmon)\n" \
	"    password=TEXT    Authentication password\n" \
	"    interval=SEC     Update interval in seconds (default: 3)\n" \
	"\n" \
	"  Examples:\n" \
	"    -O 'db:password=bmon'\n");
}

static void db_set_opts(tv_t *attrs)
{
	while (attrs) {
		if (!strcasecmp(attrs->type, "host") && attrs->value)
			c_host = attrs->value;
		else if (!strcasecmp(attrs->type, "username") && attrs->value)
			c_username = attrs->value;
		else if (!strcasecmp(attrs->type, "dbname") && attrs->value)
			c_dbname = attrs->value;
		else if (!strcasecmp(attrs->type, "password") && attrs->value)
			c_password = attrs->value;
		else if (!strcasecmp(attrs->type, "interval") && attrs->value)
			c_interval = strtol(attrs->value, NULL, 0);
		else if (!strcasecmp(attrs->type, "help")) {
			print_module_help();
			exit(0);
		}
		
		attrs = attrs->next;
	}
}

static int db_probe(void)
{
	if (!c_host)
		c_host = get_db_host();

	if (!c_username)
		c_username = get_db_user();

	if (!c_dbname)
		c_dbname = get_db_name();

	if (!c_password)
		c_password = get_db_password();

	if (c_password == NULL) {
		fprintf(stderr, "You must specify the database password\n");
		return 0;
	}

	db_conn = bdb_connect(c_host, c_username, c_password, c_dbname);
	if (db_conn == NULL)
		return 0;

	return 1;
}

static void db_shutdown(void)
{
	if (db_conn)
		bdb_shutdown(db_conn);
}

static struct output_module db_ops = {
	.om_name = "db",
	.om_draw = db_draw,
	.om_set_opts = db_set_opts,
	.om_probe = db_probe,
	.om_shutdown db_shutdown,
};

static void __init db_init(void)
{
	register_secondary_output_module(&db_ops);
}
