/*
 * in_pipe.c		       Pipe Input (Linux)
 *
 * Copyright (c) 2001-2005 Thomas Graf <tgraf@suug.ch>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

/*
 * UPDATE <timestamp>
 * NODE <name> [<source>]
 * GROUP <name>
 * ITEM <name> <handle> <parent> <parent_handle> <tmajor> <tminor> <tusage>
 * ATTR <type> <rx> <tx>
 * ATTR ...
 * END_ITEM
 * ITEM ...
 * ...
 * END_ITEM
 * END_NODE
 * END_GROUP
 * NODE ...
 * END_NODE
 * END_UPDATE
 */

#include <bmon/bmon.h>
#include <bmon/input.h>
#include <bmon/node.h>
#include <bmon/item.h>
#include <bmon/attr.h>
#include <bmon/utils.h>
#include <inttypes.h>

static int c_debug = 1;

enum {
	STATE_IDLE,
	STATE_WAIT_NODE,
	STATE_WAIT_GROUP,
	STATE_WAIT_ITEM,
	STATE_WAIT_ATTR
};

struct attr_cache {
	int			type;
	int			flags;
	b_cnt_t			rx;
	b_cnt_t			tx;
	struct attr_cache *	next;
};

struct item_cache {
	char *			name;
	char *			parent;
	struct attr_cache *	attrs;
	uint32_t		handle;
	uint32_t		parent_handle;
	int			major_attr;
	int			minor_attr;
	int			usage_attr;
	struct item_cache *	next;
};

struct group_cache {
	char *			name;
	struct item_cache *	items;
	struct group_cache *	next;
};

struct node_cache {
	char *			name;
	char *			source;
	struct group_cache *	groups;
	struct node_cache *	next;
};

struct pipe_session {
	struct node_cache *	node;
	struct group_cache *	group;
	struct item_cache *	item;
};

struct pipe_task {
	pid_t			pid;
	int			fd[2];
	char *			cmd;
	char *			args[256];
	int			state;
	char *			input;
	size_t			input_size;
	struct pipe_session	sess;
	struct node_cache *	cache;

	struct pipe_task * 	next;
};

static struct pipe_task *tasks;

static inline struct node_cache *lookup_cached_node(struct pipe_task *task,
						    const char *name)
{
	struct node_cache *node;

	for (node = task->cache; node; node = node->next)
		if (!strcmp(node->name, name))
			return node;

	node = xcalloc(1, sizeof(*node));
	node->name = strdup(name);
	node->next = task->cache;
	task->cache = node;

	return node;
}

static inline struct group_cache *lookup_cached_group(struct pipe_task *task,
						      const char *name)
{
	struct group_cache *group;

	for (group = task->sess.node->groups; group; group = group->next)
		if (!strcmp(group->name, name))
			return group;

	group = xcalloc(1, sizeof(*group));
	group->name = strdup(name);
	group->next = task->sess.node->groups;
	task->sess.node->groups = group;

	return group;
}

static inline struct item_cache *lookup_cached_item(struct pipe_task *task,
						    const char *name,
						    uint32_t handle,
						    const char *parent,
						    uint32_t parent_handle)
{
	struct item_cache *item;

	for (item = task->sess.group->items; item; item = item->next)
		if (item->handle == handle &&
		    item->parent_handle == parent_handle &&
		    !strcmp(item->name, name) &&
		    (!strcmp(parent, "<none>") ||
		     (item->parent && !strcmp(item->parent, parent))))
			return item;

	item = xcalloc(1, sizeof(*item));
	item->name = strdup(name);
	item->handle = handle;
	item->parent_handle = parent_handle;

	if (parent)
		item->parent = strdup(parent);

	item->next = task->sess.group->items;
	task->sess.group->items = item;

	return item;
}

static inline void update_attr_cache(struct pipe_task *task, int type,
				     b_cnt_t rx, b_cnt_t tx, int flags)
{
	struct attr_cache *attr;

	for (attr = task->sess.item->attrs; attr; attr = attr->next)
		if (attr->type == type)
			goto update;

	attr = xcalloc(1, sizeof(*attr));
	attr->type = type;
	attr->next = task->sess.item->attrs;
	task->sess.item->attrs = attr;

update:
	attr->rx = rx;
	attr->tx = tx;
	attr->flags = flags;
}

void new_cmd(const char *opts)
{
	int i = 0;
	struct pipe_task *task;
	char *p, *t = strdup(opts);

	task = xcalloc(1, sizeof(*task));

	p = strchr(t, ' ');
	if (p) {
		*p = '\0';
		p++;
	}

	task->cmd = t;

	if (!p || !*p)
		goto link;

	task->args[i++] = t;
	task->args[i++] = PACKAGE_VERSION;
	task->args[i++] = p;
	
	for (; i <= 255; i++) {
		p = strchr (p, ' ');
		if (p) {
			*p = '\0';
			p++;
			task->args[i] = p;
		} else
			break;
	}

	task->args[i] = NULL;

link:
	task->next = tasks;
	tasks = task;
}

static int start_task(struct pipe_task *task)
{
	if (pipe(task->fd) < 0) {
		xwarn("Could not create pipe: %s\n", strerror(errno));
		return -1;
	}
	
	task->pid = vfork();
	if (task->pid < 0) {
		xwarn("Could not fork: %s\n", strerror(errno));
		return -1;
	} else if (task->pid) {
		close(task->fd[1]);

		if (fcntl(task->fd[0], F_SETFL, O_NONBLOCK) < 0) {
			xwarn("Could not set pipe non-blocking: %s\n",
			      strerror(errno));
			return -1;
		}

		return 0;
	}

	close(task->fd[0]);

	if (dup2(task->fd[1], 1) < 0)  
		quit("Can't dup2 pipe to STDOUT: %s\n", strerror(errno));

	if (execve(task->cmd, task->args, NULL) < 0)
		quit("execve failed: %s\n", strerror(errno));
	
	_exit(0);
}

static inline void reset_state(struct pipe_task *task)
{
	xwarn("Reseting state for task %s\n", task->cmd);
	memset(&task->sess, 0, sizeof(task->sess));
	task->state = STATE_IDLE;
}

static inline void new_session(struct pipe_task *task, const char *date)
{
	memset(&task->sess, 0, sizeof(task->sess));

	task->state = STATE_WAIT_NODE;
}

static inline void new_node(struct pipe_task *task,
			    const char *name, const char *source)
{
	struct node_cache *node;

	task->sess.node = node = lookup_cached_node(task, name);

	if (source && *source) {
		if (node->source) {
			if (strcmp(source, node->source)) {
				free(node->source);
				node->source = strdup(source);
			}
		} else
			node->source = strdup(source);
	}
	
	task->state = STATE_WAIT_GROUP;
}

static inline void new_group(struct pipe_task *task, const char *name)
{
	if (!task->sess.node)
		BUG();

	task->sess.group = lookup_cached_group(task, name);
	task->state = STATE_WAIT_ITEM;
}

static inline void new_item(struct pipe_task *task, const char *name,
			    const char *handle, const char *parent,
			    const char *parent_handle, const char *major,
			    const char *minor, const char *usage)
{
	uint32_t ihandle = strtoul(handle, NULL, 0);
	uint32_t iphandle = strtoul(parent_handle, NULL, 0);
	int maj_type, min_type, usage_type;
	
	if (!task->sess.node || !task->sess.group)
		BUG();

	maj_type = attr_name2type(major);
	if (maj_type == INT_MAX) {
		xwarn("Unknown attribute type \"%s\" for major attribute\n",
		      major);
		goto reset;
	}

	min_type = attr_name2type(minor);
	if (min_type == INT_MAX) {
		xwarn("Unknown attribute type \"%s\" for minor attribute\n",
		      minor);
		goto reset;
	}

	usage_type = attr_name2type(usage);
	if (usage_type == INT_MAX) {
		xwarn("Unknown attribute type \"%s\" for usage attribute\n",
		      usage);
		goto reset;
	}

	task->sess.item = lookup_cached_item(task, name, ihandle,
					     parent, iphandle);

	task->sess.item->major_attr = maj_type;
	task->sess.item->minor_attr = min_type;
	task->sess.item->usage_attr = usage_type;
	
	task->state = STATE_WAIT_ATTR;

	return;

reset:
	reset_state(task);
}

static inline void end_group(struct pipe_task *task)
{
	if (task->sess.group && c_debug)
		fprintf(stderr, "Finished group\n");

	task->sess.group = NULL;
	task->state = STATE_WAIT_GROUP;
}

static inline void end_item(struct pipe_task *task)
{
	if (task->sess.item && c_debug)
		fprintf(stderr, "Finished item\n");

	task->sess.item = NULL;
	task->state = STATE_WAIT_ITEM;
}

static inline void new_attr(struct pipe_task *task, const char *name,
			    const char *rx, const char *tx)
{
	int type;
	b_cnt_t irx, itx;

	if (!task->sess.node || !task->sess.group || !task->sess.item)
		BUG();

	type = attr_name2type(name);
	if (type == INT_MAX)
		goto ignore;

	if (sscanf(rx, "%" SCNu64, &irx) != 1 ||
	    sscanf(tx, "%" SCNu64, &itx) != 1)
		goto ignore;

	update_attr_cache(task, type, irx, itx, RX_PROVIDED|TX_PROVIDED);
	return;

ignore:
	xwarn("Ignoring attribute update for type \"%s\" item \"%s\" " \
	      "node \"%s\" on task \"%s\"\n", name, task->sess.item->name,
	      task->sess.node->name, task->cmd);
}

static inline void end_session(struct pipe_task *task)
{
	task->state = STATE_IDLE;
}

static inline void end_node(struct pipe_task *task)
{
	task->state = STATE_WAIT_NODE;
}

#define GET_NEXT(_BUF, _DST)			\
{						\
	if (*(_BUF) == '\0')			\
		goto discard;			\
	(_DST) = (_BUF);			\
	(_BUF) = strchr((_DST), ' ');		\
	if ((_BUF)) {				\
		*(_BUF) = '\0';			\
		(_BUF)++;			\
	} else {				\
		(_BUF) = strchr((_DST), '\0');	\
		if (!(_BUF))			\
			BUG();			\
	}					\
}

#define GET_NEXT_QUOTE(_BUF, _DST)		\
{						\
	if (*(_BUF) != '"')			\
		goto discard;			\
	(_DST) = ++(_BUF);			\
	(_BUF) = strchr((_BUF), '"');		\
	if (!(_BUF))				\
		goto discard;			\
	*(_BUF) = '\0';				\
	(_BUF)++;				\
	if (*(_BUF) == ' ') {			\
		*(_BUF) = '\0';			\
		(_BUF)++;			\
	} else if (*(_BUF) != '\0')		\
		goto discard;			\
}

#define GET_OPTIONAL(_BUF, _DST)		\
{						\
	if (*(_BUF)) {				\
		(_DST) = (_BUF);		\
		(_BUF) = strchr((_DST), ' ');	\
		if ((_BUF)) {			\
			*(_BUF) = '\0';		\
			(_BUF)++;		\
		} else {			\
			(_BUF) = strchr((_DST), '\0');	\
			if (!(_BUF))		\
			    BUG();		\
		}				\
	}					\
}

static int process_group_def(struct pipe_task *task, char *buf)
{
	char *cmd, *name, *title, *left_maj_title, *left_min_title;
	char *right_maj_title, *right_min_title;
	char *cur = buf;

	GET_NEXT(cur, cmd);
	GET_NEXT(cur, name);
	GET_NEXT_QUOTE(cur, title);
	GET_NEXT_QUOTE(cur, left_maj_title);
	GET_NEXT_QUOTE(cur, left_min_title);
	GET_NEXT_QUOTE(cur, right_maj_title);
	GET_NEXT_QUOTE(cur, right_min_title);

	if (c_debug)
		fprintf(stderr, "New group definition: %s\n", name);

	add_group_def(name, title, left_maj_title, left_min_title,
		      right_maj_title, right_min_title);

	return 0;
discard:
	xwarn("Discarding message \"%s\" for task %s\n", buf, task->cmd);
	return 0;
}

static int process_while_idle(struct pipe_task *task, char *buf)
{
	char *cmd, *date, *cur = buf;

	if (!strncasecmp(cur, "group_def ", 10))
		return process_group_def(task, buf);

	GET_NEXT(cur, cmd);
	date = cur;

	if (*date == '\0')
		goto discard;
	
	if (strcasecmp(cmd, "update"))
		goto discard;

	if (c_debug)
		fprintf(stderr, "New Update: %s\n", date);

	new_session(task, date);
	return 0;

discard:
	xwarn("Discarding message \"%s\" for task %s\n", buf, task->cmd);
	reset_state(task);
	return 0;
}

static int process_while_waiting_for_node(struct pipe_task *task, char *buf)
{
	char *cmd, *name, *source = NULL, *cur = buf;

	if (!strcasecmp(cur, "end_update")) {
		end_session(task);
		return 1; /* stop parsing */
	}

	GET_NEXT(cur, cmd);
	GET_NEXT(cur, name);
	GET_OPTIONAL(cur, source);

	if (strcasecmp(cmd, "node"))
		goto discard;

	if (c_debug)
		fprintf(stderr, "New Node: %s [%s]\n",
			name, source ? source : "NULL");

	new_node(task, name, source);
	return 0;

discard:
	xwarn("Discarding message \"%s\" for task %s\n", buf, task->cmd);
	reset_state(task);
	return 0;
}

static int process_while_waiting_for_group(struct pipe_task *task, char *buf)
{
	char *cmd, *name;
	char *cur = buf;

	if (!strcasecmp(cur, "end_node")) {
		end_node(task);
		return 0;
	}

	GET_NEXT(cur, cmd);
	GET_NEXT(cur, name);

	if (strcasecmp(cmd, "group"))
		goto discard;

	if (c_debug)
		fprintf(stderr, "New Group: %s\n", name);

	new_group(task, name);
	return 0;

discard:
	xwarn("Discarding message \"%s\" for task %s\n", buf, task->cmd);
	reset_state(task);
	return 0;
}

static int process_while_waiting_for_item(struct pipe_task *task, char *buf)
{
	char *name, *handle, *parent, *parent_handle, *maj, *min, *usage;
	char *cmd, *cur = buf;

	if (!strcasecmp(cur, "end_group")) {
		end_group(task);
		return 0;
	}

	GET_NEXT(cur, cmd);
	GET_NEXT(cur, name);
	GET_NEXT(cur, handle);
	GET_NEXT(cur, parent);
	GET_NEXT(cur, parent_handle);
	GET_NEXT(cur, maj);
	GET_NEXT(cur, min);
	GET_NEXT(cur, usage);

	if (strcasecmp(cmd, "item"))
		goto discard;

	if (c_debug)
		fprintf(stderr, "New Item: %s %s %s %s %s %s %s\n",
			name, handle, parent, parent_handle,
			maj, min, usage);

	new_item(task, name, handle, parent, parent_handle,
		 maj, min, usage);
	return 0;

discard:
	xwarn("Discarding message \"%s\" for task %s\n", buf, task->cmd);
	reset_state(task);
	return 0;
}

static int process_while_waiting_for_attr(struct pipe_task *task, char *buf)
{
	char *cmd, *name, *rx, *tx, *cur = buf;

	if (!strcasecmp(cur, "end_item")) {
		end_item(task);
		return 0;
	}

	GET_NEXT(cur, cmd);
	GET_NEXT(cur, name);
	GET_NEXT(cur, rx);
	GET_NEXT(cur, tx);

	if (strcasecmp(cmd, "attr"))
		goto discard;

	if (c_debug)
		fprintf(stderr, "New Attr: %s %s %s\n", name, rx, tx);

	new_attr(task, name, rx, tx);
	return 0;

discard:
	xwarn("Discarding message \"%s\" for task %s\n", buf, task->cmd);
	reset_state(task);
	return 0;
}

static int process_cmd(struct pipe_task *task, char *buf)
{
	if (c_debug)
		fprintf(stderr, "Processing: >%s<\n", buf);

	switch (task->state) {
		case STATE_IDLE:
			return process_while_idle(task, buf);

		case STATE_WAIT_NODE:
			return process_while_waiting_for_node(task, buf);

		case STATE_WAIT_GROUP:
			return process_while_waiting_for_group(task, buf);

		case STATE_WAIT_ITEM:
			return process_while_waiting_for_item(task, buf);

		case STATE_WAIT_ATTR:
			return process_while_waiting_for_attr(task, buf);
	}

	return 0;
}

static char *parse_input(struct pipe_task *task, char *buf, int *stop)
{
	char *end;

	*stop = 0;
	
	for (;;) {
		end = strchr(buf, '\n');
		if (!end)
			return buf;

		*end = '\0';
		end++;

		*stop = process_cmd(task, buf);

		buf = end;
		if (*buf == '\0')
			return NULL;
		else if (*stop)
			return buf;
	}

	BUG();
	return NULL;
}

static int process_input(struct pipe_task *task, char *buf)
{
	int stop = 0;
	int old_size;

	if (!task->input || !strlen(task->input)) {
		buf = parse_input(task, buf, &stop);

		if (buf == NULL)
			return stop;
	}

	old_size = task->input_size;
	task->input_size += strlen(buf);

	if (!old_size) {
		/* room for trailing \0 */
		task->input_size++; 
	}
	
	task->input = realloc(task->input, task->input_size);
	if (old_size)
		strcat(task->input, buf);
	else
		memcpy(task->input, buf, strlen(buf) + 1);

	if (stop)
		return stop;

	buf = parse_input(task, task->input, &stop);

	if (buf == NULL) {
		xfree(task->input);
		task->input = NULL;
		task->input_size = 0;
	} else {
		old_size = task->input_size;
		task->input_size -= (buf - task->input);
		memmove(task->input, buf, task->input_size);
	}

	return stop;
}

void pipe_handle(void)
{
	int n, status;
	char buf[1024];
	struct pipe_task *t;

	for (t = tasks; t; t = t->next) {
		for (;;) {
			memset(buf, 0, sizeof(buf));

			n = read(t->fd[0], buf, sizeof(buf) - 1);

			if (n < 0) {
				if (errno == EAGAIN)
					break;
				else {
					xwarn("Error while reading from " \
					      "pipe: %s\n", strerror(errno));
					break;
				}
			} else if (n > 0) {
				/* A non-zero return value means that a
				 * complete update session has been parsed
				 * and processed, we need to return and give
				 * the core system a chance to take the new
				 * input into account and have the updated
				 * flags be reset. */
				if (process_input(t, buf))
					break;
			} else
				break;
		}
	}

	while (waitpid(0, &status, WNOHANG) > 0);
}

void pipe_read(void)
{
	struct pipe_task *t;

	for (t = tasks; t; t = t->next) {
		struct node_cache *node;
		
		for (node = t->cache; node; node = node->next) {
			struct node *real_node;
			struct group_cache *group;

			if (!strcasecmp(node->name, "<local>"))
				real_node = get_local_node();
			else
				real_node = lookup_node(node->name, 1);

			if (real_node == NULL)
				continue;

			for (group = node->groups; group; group = group->next) {
				struct item_group *real_group;
				struct item_cache *item;

				real_group = get_group(real_node, group->name, 1);
				if (real_group == NULL)
					continue;

				for (item = group->items; item; item = item->next) {
					struct item *real_item, *p = NULL;
					struct attr_cache *attr;
					int level = 0;

					if (strcasecmp(item->parent, "<none>")) {
						p = search_item(real_group, item->parent,
								item->parent_handle, NULL);

						if (p == NULL)
							continue;

						level = p->i_level + 1;
					}

					real_item = lookup_item(real_node,
								real_group,
								item->name,
								item->handle, p);

					if (real_item == NULL)
						continue;

					real_item->i_major_attr = item->major_attr;
					real_item->i_minor_attr = item->minor_attr;
					real_item->i_usage_attr = item->usage_attr;
					real_item->i_level = level;

					for (attr = item->attrs; attr; attr = attr->next)
						update_attr(real_item, attr->type,
							    attr->rx, attr->tx, attr->flags);

					notify_update(real_item, NULL);
					increase_lifetime(real_item, 1);
				}
			}
		}
	}
}

int pipe_start(void)
{
	struct pipe_task *t;

	for (t = tasks; t; t = t->next)
		if (start_task(t) < 0)
			return 0;
	
	return 1;
}
