/*
 * statefile.c			Statefile
 *
 * Copyright (c) 2001-2005 Thomas Graf <tgraf@suug.ch>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <bmon/bmon.h>
#include <bmon/conf.h>
#include <bmon/item.h>
#include <bmon/node.h>
#include <bmon/statefile.h>
#include <bmon/utils.h>

struct xdata {
	FILE *fd;
	struct item *it;
};

static void dump_attr(struct stat_attr *a, void *arg)
{
	struct xdata *x = arg;
	struct item *it = x->it;
	FILE *fd = x->fd;
	rate_cnt_t rx_rate = attr_get_rx_rate(a);
	rate_cnt_t tx_rate = attr_get_tx_rate(a);
	b_cnt_t rx_sum = attr_get_rx(a);
	b_cnt_t tx_sum = attr_get_tx(a);
	b_cnt_t rx_in = attr_get_rx_last(a);
	b_cnt_t tx_in = attr_get_tx_last(a);

	struct item *parent = NULL;

	if (it->i_flags & ITEM_FLAG_IS_CHILD)
		parent = get_item(get_local_node(), it->i_parent);

	fprintf(fd, "%-14s\t%-14s\t%08X\t%08X\t%-14s\t%08X\t%08X\t%016"
		PRIX64 "\t%016" PRIX64 "\t%016" PRIX64 "\t%016" PRIX64 "\t" \
		    "%016" PRIX64 ":%016" PRIX64 "\n",
		it->i_name,
		parent ? parent->i_name : "<none>",
		htonl(it->i_handle),
		parent ? htonl(parent->i_handle) : 0,
		type2name(a->a_type),
		htonl(rx_rate),
		htonl(tx_rate),
		xhtonll(rx_sum),
		xhtonll(tx_sum),
		xhtonll(rx_in),
		xhtonll(tx_in),
		xhtonll(a->a_updated.tv_sec),
		xhtonll(a->a_updated.tv_usec));
}

static void dump_item(FILE *fd, struct item *it)
{
	struct xdata x = {
		.fd = fd,
		.it = it,
	};

	foreach_attr(it, dump_attr, &x);
}


static void handle_child(struct item *it, void *arg)
{
	FILE *fd = arg;

	dump_item(fd, it);
	foreach_child(get_local_node(), it, handle_child, fd);
}

static void write_per_item(struct item *it, void *arg)
{
	FILE *fd = arg;

	if (it->i_flags & ITEM_FLAG_IS_CHILD)
		return;

	dump_item(fd, it);
	foreach_child(get_local_node(), it, handle_child, fd);
}

void write_statefile(const char *path, int lock)
{
	FILE *fd;
	struct node *local_node;

	fd = fopen(path, "w");
	if (fd == NULL) {
		xwarn("Could not open statefile \"%s\": %s\n",
		      path, strerror(errno));
		return;
	}

	if (lock && flock(fileno(fd), LOCK_EX | LOCK_NB) < 0) {
		xwarn("Could not lock statefile \"%s\": %s\n",
		      path, strerror(errno));
		return;
	}

	fprintf(fd, "# <name> <parent> <handle> <attribute> <rx_rate> " \
		    "<tx_rate> <rx_total> <tx_total> <rx_in> <tx_in> " \
		    "<ts_secs>:<ts_usecs>\n");

	local_node = get_local_node();

	if (local_node)
		foreach_item(local_node, write_per_item, fd);

	if (lock)
		flock(fileno(fd), LOCK_UN | LOCK_NB);

	fclose(fd);
}

int open_statefile(const char *path, int lock, FILE **fd)
{
	*fd = fopen(path, "r");
	if (fd == NULL) {
		xwarn("Could not open statefile \"%s\": %s\n",
		    path, strerror(errno));
		return -ENOENT;
	}

	if (lock && flock(fileno(*fd), LOCK_EX | LOCK_NB) < 0)
		return -EAGAIN;

	return 0;
}

struct sf_token *read_next_token(FILE *fd)
{
	int ret, t;
	char buf[1024];
	struct sf_token *tok = NULL;
	char attr_type[32];

skip:
	if (!fgets(buf, sizeof(buf), fd))
		goto errout;

	if (*buf == '#' || *buf == '\n' || *buf == '\r')
		goto skip;

	tok = xcalloc(1, sizeof(*tok));

	ret = sscanf(buf, "%s\t%s\t%08X\t%08X\t%s\t%08X\t%08X\t%016" PRIX64
		     "\t%016" PRIX64 "\t%016" PRIX64 "\t%016" PRIX64 "\t%016"
		     PRIX64 ":%016" PRIX64,
		     tok->st_name, tok->st_parent, &tok->st_handle,
		     &tok->st_parent_handle, attr_type, &tok->st_rx_rate,
		     &tok->st_tx_rate, &tok->st_rx_sum, &tok->st_tx_sum,
		     &tok->st_rx_last, &tok->st_tx_last,
		     &tok->st_ts.tv_sec, &tok->st_ts.tv_usec);

	if (ret != 10)
		goto errout;
	
	t = name2type(attr_type);
	if (t == INT_MAX)
		goto errout;

	tok->st_handle = ntohl(tok->st_handle);
	tok->st_parent_handle = ntohl(tok->st_parent_handle);
	tok->st_attr_type =t ;
	tok->st_rx_rate = ntohl(tok->st_rx_rate);
	tok->st_tx_rate = ntohl(tok->st_tx_rate);
	tok->st_rx_sum = xntohll(tok->st_rx_sum);
	tok->st_tx_sum = xntohll(tok->st_tx_sum);
	tok->st_ts.tv_sec = xntohll(tok->st_ts.tv_sec);
	tok->st_ts.tv_usec = xntohll(tok->st_ts.tv_usec);
	
	return tok;

errout:
	if (tok)
		xfree(tok);
	return NULL;
}

void close_statefile(int lock, FILE *fd)
{
	if (lock)
		flock(fileno(fd), LOCK_UN | LOCK_NB);

	fclose(fd);
}

static void notify_after_restore(struct item *it, void *arg)
{
	increase_lifetime(it, 1);
}

int restore_from_statefile(const char *path, int lock)
{
	int err;
	FILE *fd;
	int nretries = 10;
	struct sf_token *tok;
	struct node *node;

	while (nretries--) {
		err = open_statefile(path, lock, &fd);
		if (err < 0) {
			if (err != -EAGAIN)
				return err;
		} else
			break;
	}

	node = get_local_node();

	while ((tok = read_next_token(fd))) {
		struct item *it;

		if (strcmp(tok->st_parent, "<none>"))
			it = lookup_item(node, tok->st_name, tok->st_handle,
					 NULL);
		else {
			while (stackp >= 0) {
				if (!strcmp(stack[stackp]->i_name,
					    tok->st_parent))
					goto found_parent;
				stackp--;
			}

			xwarn("Couldn't find parent \"%s\" for \"%s\"\n",
			      tok->st_parent, tok->st_name);
			goto skip;

found_parent:
			it = lookup_item(node, tok->st_name, tok->st_handle,
					 stack[stackp]);

			stackp++;
		}

		if (it == NULL)
			goto skip;

		update_attr(it, tok->st_attr_type, tok->st_rx_last,
			    tok->st_tx_last,
			    tok->st_rx_last ? RX_PROVIDED : 0 |
			    tok->st_tx_last ? TX_PROVIDED : 0);

		restore_attr(it, tok->st_attr_type, tok->st_rx_sum,
			     tok->st_tx_sum);

		stack[stackp] = it;
skip:
		xfree(tok);
	}

	foreach_item(node, &notify_after_restore, NULL);

	close_statefile(lock, fd);

	return 0;
}
