/*
 * unit.c		Units
 *
 * Copyright (c) 2001-2005 Thomas Graf <tgraf@suug.ch>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <bmon/bmon.h>
#include <bmon/node.h>
#include <bmon/conf.h>
#include <bmon/item.h>
#include <bmon/attr.h>
#include <bmon/unit.h>
#include <bmon/utils.h>

static struct unit *units;
static int unit_gen = 1;

struct unit *lookup_unit(const char *name)
{
	struct unit *u;

	for (u = units; u; u = u->u_next)
		if (!strcmp(name, u->u_name))
			return u;

	return NULL;
}

static inline b_cnt_t __divisor(struct unit *unit, int exp)
{
	b_cnt_t div;

	if (get_use_si() && unit->u_have_si)
		div = unit->u_si_divisors[exp];
	else
		div = unit->u_divisors[exp];

	return div ? div : 1;
}

static inline char * __unit(struct unit *unit, int exp)
{
	if (get_use_si() && unit->u_have_si)
		return unit->u_si_units[exp] ? unit->u_si_units[exp] : "";
	else
		return unit->u_units[exp] ? unit->u_units[exp] : "";
}

b_cnt_t get_divisor(b_cnt_t hint, struct unit *unit, char **dst_unit, int *prec)
{
	int i, yunit = get_y_unit();

	if (prec)
		*prec = 2;

	if (yunit == Y_DYNAMIC) {
		int max = unit->u_idx;
		if (get_use_si() && unit->u_have_si)
			max = unit->u_si_idx;
		for (i = max; i >= 0; i--) {
			b_cnt_t div = __divisor(unit, i);
			if (hint >= div || i == 0) {
				*dst_unit = __unit(unit, i);
				if (div == 1 && prec)
					*prec = 0;
				return div;
			}
		}
	} else {
		*dst_unit = __unit(unit, yunit);
		return __divisor(unit, yunit);
	}

	return 1;
}

inline double cancel_down(b_cnt_t l, struct unit *unit,
			  char **dst_unit, int *prec)
{
	return ((double) l / (double) get_divisor(l, unit, dst_unit, prec));
}

void add_unit(const char *name, const char *def)
{
	char *p, *next, *buf = strdup(def);
	char *current = buf;
	int idx = 0, up_si = 0;
	struct unit *u;

	p = strrchr(name, '/');
	if (p && !strcmp(p, "/si")) {
		char *parent = strdup(name);

		parent[strlen(parent)-3] = '\0';
		u = lookup_unit(parent);
		if (u == NULL)
			quit("Could not find parent unit \"%s\" for SI " \
			     "unit \"%s\"\n", parent, name);
		up_si = 1;
		memset(u->u_si_divisors, 0, sizeof(u->u_si_divisors));
		memset(u->u_si_units, 0, sizeof(u->u_si_units));
		goto update_si;
	}

	u = lookup_unit(name);

	if (u == NULL)
		u = xcalloc(1, sizeof(*u));

	if (u->u_name)
		xfree(u->u_name);
	u->u_name = strdup(name);

	memset(u->u_divisors, 0, sizeof(u->u_divisors));
	memset(u->u_units, 0, sizeof(u->u_units));

update_si:
	do {
		next = strchr(current, ',');

		if (next) {
			*next = '\0';
			++next;
		}

		p = strchr(current, ':');
		if (!div)
			quit("Parse error near \"%s\", no divisor found\n",
			     current);

		*p = '\0';
		p++;

		if (idx >= MAX_DIVISOR)
			quit("Exceeding MAX_DIVISOR(%d)... HARRR\n",
			     MAX_DIVISOR);

		if (up_si)
			u->u_si_divisors[idx] = strtod(current, NULL);
		else
			u->u_divisors[idx] = strtod(current, NULL);

		if (*p) {
			if (up_si)
				u->u_si_units[idx] = strdup(p);
			else
				u->u_units[idx] = strdup(p);
		}

		idx++;

		current = next;
	} while (next);

	if (up_si)
		u->u_si_idx = (idx - 1);
	else
		u->u_idx = (idx - 1);

	if (!u->u_id) {
		u->u_id = unit_gen++;
		u->u_next = units;
		units = u;
	}

#ifdef DEBUG
	{
		int i;

		DBG(1, "[DBG] %s unit (%s/%d):",
		    up_si ? "Updated SI" : "New", u->u_name, u->u_id);
		for (i = 0; i < idx; i++)
			DBG(1, "%" PRIu64 "=%s;",
			    up_si ? u->u_si_divisors[i] : u->u_divisors[i],
			    up_si ? u->u_si_units[i] : u->u_units[i]);
		DBG(1, "\n");
	}
#endif

	free(buf);
}
